import os
import os.path
import traceback
from logging import Logger

from ..settings import logger as mylogger
from .character import Character
from .prefixmap import PrefixMap
from .oto import Oto


default_logger = mylogger.get_logger(__name__, False)


class VoiceBank:
    '''VoiceBank
    UTAUの音源データを扱います。
    
    Attributes
    ----------
    dirpath: str
        音源のルートパス

    character: Character
        character.txt

    oto: Oto
        oto.ini

    prefix: PrefixMap
        prefix.map
    '''

    _dirpath: str
    _character: Character
    _oto: Oto
    _prefix: PrefixMap

    @property
    def dirpath(self) -> str:
        return self._dirpath

    @property
    def character(self) -> Character:
        return self._character

    @property
    def oto(self) -> Oto:
        return self._oto

    @property
    def prefix(self) -> PrefixMap:
        return self._prefix

    def __init__(self, dirpath: str, *, logger:Logger = None):
        '''
        Parameters
        ----------
        dirpath: str
            音源のルートパス

        Raises
        ------
        FileNotFoundErrod
            指定したフォルダが見つからなかったとき

        ValueError
            指定したフォルダが音源フォルダではなかったとき
        '''
        self.logger = logger or default_logger
        if not VoiceBank.is_utau_voicebank(dirpath):
            self.logger.error("{} is not utau voicebanks".format(dirpath))
            raise ValueError("{} is not utau voicebanks".format(dirpath))
        self._dirpath = dirpath
        try:
            self._character = Character(dirpath)
            self.logger.info("character.txt is loaded. VBName:{}".format(self._character.name))
        except Exception as e:
            self.logger.warn(traceback.format_exception_only(type(e), e)[0].rstrip('\n'))
            self._character = Character()
            
        try:
            self._oto = Oto(dirpath)
            self.logger.info("oto.ini is loaded.files {}, records {}".format(self._oto.files(), self._oto.records()))
        except Exception as e:
            self.logger.warn(traceback.format_exception_only(type(e), e)[0].rstrip('\n'))

        try:
            self._prefix = PrefixMap(dirpath)
            self.logger.info("prefix.map is loaded")
        except Exception as e:
            self.logger.warn(traceback.format_exception_only(type(e), e)[0].rstrip('\n'))
            self._prefix = PrefixMap()

    @staticmethod
    def is_utau_voicebank(dirpath: str, *, logger:Logger = None) -> bool:
        '''
        | 渡されたパスがUTAU音源フォルダか判定する。
        | character.txt、oto.ini、.wavのいずれかがあればUTAU音源フォルダと判定する。

        Parameters
        ----------
        dirpath: str
            音源のルートパス

        Returns
        -------
        is_utau_voicebank: bool
            
        Raises
        ------
        FileNotFoundErrod
            指定したフォルダが見つからなかったとき

        '''
        logger = logger or default_logger
        if not os.path.isdir(dirpath):
            logger.error("{} is not found or not directory".format(dirpath))
            raise FileNotFoundError("{} is not found or not directory".format(dirpath))
        logger.info("{} is found. checking directory".format(dirpath))
        files: list = os.listdir(dirpath)
        if "character.txt" in files:
            logger.info("{} is found.".format("character.txt"))
            return True
        elif "oto.ini" in files:
            logger.info("{} is found.".format("oto.ini"))
            return True
        elif ".wav:" in ":".join(files).lower():
            logger.info("{} is found.".format("wav file"))
            return True
        elif ":".join(files).lower().endswith(".wav"):
            logger.info("{} is found.".format("wav file"))
            return True
        else:
            return False




import os.path
import hashlib

from typing import Tuple

import numpy as np
import PyRwu.pitch

from .Note import Note
from ..voicebank import VoiceBank
from .. import settings

class RenderNote:
    '''
    UTAUのNoteをresamplerやwavtoolに渡せるパラメータにしたもの。

    Attributes
    ----------
    input_path: str
        入力する原音のフルパス。

    cache_path: str
        resampが出力する中間ファイルのパス

    output_path: str
        wavtoolが出力する最終ファイルのパス

    target_tone: str

        | 音高名(A4=440Hz)。
        | 半角上げは#もしくは♯ 
        | 半角下げはbもしくは♭で与えられます。

    velocity: int
        子音速度

    flags: str, default ""
        フラグ(省略可)

    offset: float, default 0
        入力ファイルの読み込み開始位置(ms)

    target_ms: float, default 0

        | 出力ファイルの長さ(ms)
        | UTAUでは通常50ms単位に丸めた値が渡される。

    fixed_ms: float, default 0
        offsetからみて通常伸縮しない長さ(ms)

    end_ms: float, default 0
        | 入力ファイルの読み込み終了位置(ms)(省略可 default:0)
        | 正の数の場合、ファイル末尾からの時間
        | 負の数の場合、offsetからの時間

    intensity: int, default 100
        音量。0～200(省略可)

    modulation: int, default 0
        モジュレーション。0～200(省略可)

    tempo: str, default "!120"

        | ピッチのテンポ
        | 数字の頭に!がついた文字列

    pitchbend: str, default ""

        | ピッチベンド。(省略可)
        | -2048～2047までの12bitの2進数をbase64で2文字の文字列に変換し、
        | 同じ数字が続く場合ランレングス圧縮したもの

    stp : float
        入力wavの先頭のオフセットをmsで指定する。

    output_ms: float
        wavtoolが出力する長さ

    
    envelope : list
        エンベロープのパターンは以下のいずれかです。
            >>> p1 p2
            >>> p1 p2 p3 v1 v2 v3 v4
            >>> p1 p2 p3 v1 v2 v3 v4 ove
            >>> p1 p2 p3 v1 v2 v3 v4 ove p4
            >>> p1 p2 p3 v1 v2 v3 v4 ove p4 p5 v5
        | p1,p2,p3,p4,p5,ove : float
        | v1,v2,v3,v4,v5 : int

    '''
    _input_path: str
    _output_path: str
    _target_tone: str
    _velocity: int
    _flags: str
    _offset: float
    _target_ms: int
    _fixed_ms: float
    _end_ms: float
    _intensity: int
    _modulation: int
    _tempo: str
    _pitchbend: str
    _stp: float
    _envelope: str
    _cache_path: str
    _output_ms: float
    _require_resamp: bool = True
    _direct: bool = False

    @property
    def input_path(self) -> str:
        return self._input_path
    
    @property
    def output_path(self) -> str:
        return self._output_path

    @property
    def target_tone(self) -> str:
        return self._target_tone
    
    @property
    def velocity(self) -> int:
        return self._velocity

    @property
    def flags(self) -> str:
        return self._flags
    
    @property
    def offset(self) -> float:
        return self._offset
    
    @property
    def target_ms(self) -> int:
        return self._target_ms
    
    @property
    def fixed_ms(self) -> float:
        return self._fixed_ms
    
    @property
    def end_ms(self) -> float:
        return self._end_ms
    
    @property
    def intensity(self) -> int:
        return self._intensity
    
    @property
    def modulation(self) -> int:
        return self._modulation

    @property
    def tempo(self) -> str:
        return self._tempo
    
    @property
    def pitchbend(self) -> str:
        return self._pitchbend
    
    @property
    def stp(self) -> float:
        return self._stp

    @property
    def envelope(self) -> str:
        return self._envelope
    
    @property
    def cache_path(self) -> str:
        return self._cache_path
    
    @property
    def output_ms(self) -> float:
        return self._output_ms

    @property
    def require_resamp(self) -> bool:
        return self._require_resamp

    @property
    def direct(self) -> bool:
        return self._direct

    def __init__(self, note: Note, vb: VoiceBank, cachedir: str, output: str, mode2: bool = True):
        '''
        Parameters
        ----------
        note: Note
            | 変換するノート。
            | noteとnote.nextは事前にapply_otoを実施しておくこと

        vb: VoiceBank
            音源

        cachedir: str
            resampが中間ファイルを出力するフォルダ

        output: str
            wavtoolがwavを出力するフォルダ

        mode2: bool True
            | ピッチデータの扱い。
            | Trueの場合、PBS,PBY,PBM,PBW,Vibratoを解釈
            | Falseの場合、Pitches,PBStartを解釈
        '''
        self._input_path = os.path.join(vb.dirpath, note.atFileName.value)
        self._output_path = output
        self._target_tone = note.notenum.get_tone_name()
        self._velocity = note.velocity.value
        self._flags = note.flags.value
        if note.next is None:
            self._output_ms = note.msLength + note.atPre.value
        else:
            self._output_ms = note.msLength + note.atPre.value + note.next.atOve.value - note.next.atPre.value
        self._target_ms = (round((self._output_ms + note.atStp.value) / 50) + 1 ) * 50
        if  vb.oto.haskey(note.atAlias.value):
            self._offset = vb.oto[note.atAlias.value].offset
            self._fixed_ms = vb.oto[note.atAlias.value].consonant
            self._end_ms = vb.oto[note.atAlias.value].blank
        else:
            self._offset = 0
            self._fixed_ms = 0
            self._end_ms = 0
            self._require_resamp = False
        self._intensity = note.intensity.value
        self._modulation = note.modulation.value
        self._tempo = "!{:.2f}".format(note.tempo.value)
        if note.lyric.value != "R":
            self._pitchbend = self._get_pitches(note, mode2)
        else:
            self._pitchbend = ""
        self._stp = note.atStp.value
        if note.envelope.hasValue:
            if "%" in note.envelope.value:
                self._envelope = note.envelope.value.replace("%", str(note.atOve)).replace(","," ")
            else:
                self._envelope = note.envelope.value.replace(","," ") + " " +str(note.atOve)
        else:
            self._envelope = settings.DEFAULT_ENV.replace("%", str(note.atOve))
        self._cache_path = os.path.join(cachedir, "{}_{}_{}_{}.wav".format(note.num.value[1:],
                                                note.atAlias.value.replace(" ","+"),
                                                note.notenum.get_tone_name(),
                                                self._get_cache_hash(note)))
        if note.direct.value:
            self._require_resamp = False
            self._direct = True


    def _get_cache_hash(self, note: Note) -> str:
        '''
        resampが使用する各パラメータを使用して、ハッシュ値を生成します。

        Parameters
        ----------
        note: Note
            | 変換するノート。
            | noteとnote.nextは事前にapply_otoを実施しておくこと

        Returns
        -------
        hash: str
            6桁のハッシュ文字

        '''
        return hashlib.md5("{}_{}_{}_{}_{}_{}_{}_{}".format(note.pre.value,
                                                            note.stp.value,
                                                            note.velocity.value,
                                                            note.flags.value,
                                                            note.intensity.value,
                                                            note.modulation.value,
                                                            note.tempo.value,
                                                            self._pitchbend,
                                                            ).encode()).hexdigest()[:6]

    def _get_pitches(self, note: Note, mode2: bool) -> str:
        '''
        | mode2の値に応じてピッチ列を返します。
        | Falseの場合、Pitches,PBStartを解釈
        | Trueの場合、PBS,PBY,PBM,PBW,Vibrato,next.PBS,next.PBY.next.PBM,next.PBW,prev.PBS,prev.PBY,prev.PBM,prev.PBW,prev.Vibratoを解釈

        Parameters
        ----------
        note: Note
            | 変換するノート。
            | noteとnote.nextは事前にapply_otoを実施しておくこと

        mode2: bool True
            | ピッチデータの扱い。
            | Trueの場合、PBS,PBY,PBM,PBW,Vibratoを解釈
            | Falseの場合、Pitches,PBStartを解釈

        Returns
        -------
        pitchbend: str
            | -2048～2047までの12bitの2進数をbase64で2文字の文字列に変換し、
            | 同じ数字が続く場合ランレングス圧縮したもの

        '''
        t: np.ndarray = PyRwu.pitch.getPitchRange(self._tempo, self._target_ms, 44100)
        offset: float = note.atPre.value + note.atStp.value
        if mode2:
            base_pitches: np.ndarray = self._get_base_pitches(note, t)
            if note.prev is not None and note.prev.lyric.value != "R":
                base_pitches += self._interp_pitches(note.prev, t, offset - note.prev.msLength)
                base_pitches += self._get_vibrato_pitches(note.prev, t, offset - note.prev.msLength)
            base_pitches += self._interp_pitches(note, t, offset)
            base_pitches += self._get_vibrato_pitches(note, t, offset)
            if note.next is not None and note.next.lyric.value != "R":
                base_pitches += self._interp_pitches(note.next, t, offset + note.msLength)
        else:
            base_pitches: np.ndarray = np.zeros_like(t, dtype = np.int16)
            start_ms: float = offset + note.pbStart.value #pbstartは負の数
            start: int = np.where(t>= start_ms)[0][0]
            end: int = start + len(note.pitches.value)
            print(note.pbStart.value)
            print(offset)
            print(start)
            print(end)
            if start < end:
                base_pitches[start:end] = np.array(note.pitches.value)
        return self.encodeRunLength(self.encodeBase64(base_pitches))

    def _get_base_pitches(self, note: Note, t:np.ndarray) -> np.ndarray:
        '''
        noteとその前後のpbsとnotenumを使って、基準ピッチ列を返します。
        
        Parameters
        ----------
        note: Note
            | 変換するノート。
            | noteとnote.nextは事前にapply_otoを実施しておくこと

        t: np.ndarray
            ピッチ点の時間列

        Returns
        -------
        base_pitch: np.ndarray
            | 基準ピッチ列
            | prev.pbs.timeより前の点 = 0
            | prev.pbs.time→pbs.timeの点 = (prev.notenum - notenum) * 100
            | pbs.time → next.pbs.timeの点 = 0
            | next.pbs.timeより後ろの点 = (next.notenum - notenum) * 100
        '''
        base_pitches: np.ndarray = np.zeros_like(t, dtype = np.int16)
        offset: float = note.atPre.value + note.atStp.value
        start: int
        end: int
        if note.prev is not None and note.prev.lyric.value != "R":
            prev_offset: float = offset - note.prev.msLength
            if note.prev.pbs.time + prev_offset < 0:
                start = 0
            else:
                start = np.where(t >= note.prev.pbs.time + prev_offset)[0][0]
            if t[0]<=note.pbs.time + offset:
                end = np.where(t < note.pbs.time + offset)[0][-1]
            else:
                end = 0
            if start < end:
                base_pitches[start:end] = (note.prev.notenum.value - note.notenum.value) * 100
        if note.next is not None and note.next.lyric.value != "R":
            next_offset: float = offset + note.msLength
            if t[-1] >= note.next.pbs.time + next_offset:
                start = np.where(t >= note.next.pbs.time + next_offset)[0][0]
                base_pitches[start:] = (note.next.notenum.value - note.notenum.value) * 100
        return base_pitches

    def _interp_pitches(self, note: Note, t:np.ndarray, offset:float) -> np.ndarray:
        '''
        noteのpbs,pby,pbw,pbmとnote.prev.notenumを使ってピッチ列を返します。
        
        Parameters
        ----------
        note: Note
            | 変換するノート。
            | noteとnote.nextは事前にapply_otoを実施しておくこと

        t: np.ndarray
            ピッチ点の時間列

        offset: float
            pbs.timeなどの時間のずれ

        Returns
        -------
        pitches: np.ndarray
            ピッチ列
        '''
        pitches: np.ndarray = np.zeros_like(t, dtype = np.int16)

        if not note.pbs.hasValue:
            return pitches

        x, y, mode = self._get_interp_base(note, offset)
            
        start: int
        end: int
        for i in range(len(x)):
            if i==0:
                continue
            if t[-1] <= x[i]:
                continue
            start, end, cycle, height, phase = self._get_interp_param(x, y, t, i)
            if start >= end:
                continue
            if(mode[i-1] == ""):
                #cos(-pi) → cos(0)の補完
                pitches[start:end+1] = self._interp_default(cycle, height, phase, y[i-1])
            elif (mode[i-1] == "s"):
                #線形補完
                pitches[start:end+1] = self._interp_s(cycle, height, phase, y[i-1])
            elif (mode[i-1] == "r"):
                #sin(0) → sin(pi/2)の補完
                pitches[start:end+1] = self._interp_r(cycle, height, phase, y[i-1])
            elif (mode[i-1] == "j"):
                #cos(0) → cos(pi/2)の補完
                pitches[start:end+1] = self._interp_j(cycle, height, phase, y[i-1])

        return pitches

    @staticmethod
    def _interp_default(cycle, height, phase, offset):
        return ((np.cos(np.pi / cycle * phase - np.pi) + 1) * height / 2 + offset)
    
    @staticmethod
    def _interp_s(cycle, height, phase, offset):
        return height / cycle * phase + offset
    
    @staticmethod
    def _interp_r(cycle, height, phase, offset):
        return np.sin(np.pi / cycle / 2 * phase) * height + offset
    
    @staticmethod
    def _interp_j(cycle, height, phase, offset):
        return (-np.cos(np.pi / cycle / 2 * phase)+1) * height + offset

    def _get_interp_param(self, x: np.ndarray, y: np.ndarray, t: np.ndarray, i: int) -> Tuple[int, int, float, float, np.ndarray]:
        '''
        x[i-1]～x[i]の間のピッチパラメータを求めるための諸元を求めます。
        
        Parameters
        ----------
        x: np.ndarray
            ポルタメントの時間間隔

        y: np.ndarray
            ポルタメントの音高

        t: np.ndarray
            ピッチ点の時間列

        i: xのindex

        Returns
        -------
        start: int
            ピッチを補完する範囲の始点

        end: int
            ピッチを補完する範囲の終点

        cycle: float
            補完範囲の時間

        height: float
            補完範囲の音高差

        phase: np.ndarray
            t[start:end+1]をx[i-1]からの経過時間に変換したもの
        '''
        start: int = np.where(t>=x[i-1])[0][0]
        if t[0]<x[i]:
            end: int = np.where(t<x[i])[0][-1]
        else:
            end: int = 0
        cycle: float = x[i] - x[i-1]
        height: int = y[i] - y[i-1]
        phase: np.ndarray = t[start:end+1] - x[i-1]
        return start, end, cycle, height, phase

    def _get_interp_base(self, note: Note, offset:float) -> Tuple[np.ndarray, np.ndarray, list]:
        '''
        noteのpbs,pby,pbw,pbmとnote.prev.notenumを使ってinterpに使用するx,y,modeを求める
        
        Parameters
        ----------
        note: Note
            | 変換するノート。
            | noteとnote.nextは事前にapply_otoを実施しておくこと

        offset: float
            pbs.timeなどの時間のずれ

        Returns
        -------
        x: np.ndarray
            ポルタメントの時間間隔

        y: np.ndarray
            ポルタメントの音高

        mode: list
            ピッチの補完方法
        '''
        x:np.ndarray = np.zeros(len(note.pbw.value) + 1)
        for i in range(len(note.pbw.value)):
            x[i+1] = x[i] + note.pbw.value[i]
        x = x + note.pbs.time + offset
        y: np.ndarray = np.zeros_like(x)
        if note.prev is not None and note.prev.lyric.value != "R":
            y[0] = (note.prev.notenum.value - note.notenum.value) * 100
        else:
            y[0] = note.pbs.height * 10
        for i in range(len(note.pby.value)):
            y[i+1] = note.pby.value[i] * 10
        mode: list = (note.pbm.value + [""]*(len(x)-1))[:len(x)-1]

        return x, y, mode
    
    def _get_vibrato_pitches(self, note:Note, t:np.ndarray, offset:float) -> np.ndarray:
        '''
        noteのpbs,pby,pbw,pbmとnote.prev.notenumを使ってピッチ列を返します。
        
        Parameters
        ----------
        note: Note
            | 変換するノート。
            | noteとnote.nextは事前にapply_otoを実施しておくこと

        t: np.ndarray
            ピッチ点の時間列

        offset: float
            時間のずれ

        Returns
        -------
        pitches: np.ndarray
            ピッチ列
        '''
        pitches: np.ndarray = np.zeros_like(t, dtype = np.int16)
        if not note.vibrato.hasValue:
            return pitches
        start_ms: float = offset + note.msLength * (100 - note.vibrato.length) / 100
        end_ms: float = offset + note.msLength
        start: int = np.where(t>=start_ms)[0][0]
        if t[0] < end_ms:
            end: int = np.where(t<end_ms)[0][-1]
        else:
            end: int =0
        if start >= end:
            return pitches
                
        phase: np.ndarray = t[start:end + 1] - start_ms
        phase_offset: float = 2 * np.pi * note.vibrato.phase /100

        fade = self._get_vibrato_fade(note, phase)

        height: np.ndarray = fade * int(note.vibrato.depth)
        pitches[start:end+1] = np.round((np.sin(2 * np.pi / note.vibrato.cycle * phase + phase_offset) + note.vibrato.height/100) * height)
        return pitches

    def _get_vibrato_fade(self, note:Note, t:np.ndarray) -> np.ndarray:
        vibrato_ms: float = note.msLength * note.vibrato.length / 100
        fadeintime: float = vibrato_ms * note.vibrato.fadeInTime / 100
        fadeout_start_ms: float = vibrato_ms - vibrato_ms * note.vibrato.fadeOutTime / 100
        fade:np.ndarray = np.ones_like(t,dtype=np.float64)
        for i in range(len(t)):
            if t[i] <= fadeintime and fadeintime != 0:
                fade[i] =  t[i] / fadeintime
            elif t[i] >= fadeout_start_ms:
                fade[i] =  1 - ((t[i] - fadeout_start_ms) / (vibrato_ms - fadeout_start_ms))
        return fade

    @staticmethod
    def encodeBase64Core(value: int) -> str:
        '''
        0～63の数値を受け取り、1文字のstrを返す。

        Parameters
        ----------
        value: int

        Returns
        -------
        result: str
        '''
        if value < 26: #A-Z
            return chr(value + ord("A"))
        elif value < 52:
            return chr(value + ord("a") - 26)
        elif value < 62:
            return chr(value + ord("0") - 52)
        elif value == 62:
            return "+"
        else:
            return "/"

    @staticmethod
    def encodeBase64(values: np.ndarray) -> list:
        '''
        -2048 ～ 2047の数字を受け取り、文字列を返します。

        Parameters
        ----------
        values: np.ndarray
            -2048 ～ 2047のint列

        Returns
        -------
        result: list
            base64にエンコードした2桁の文字列のリスト
        '''
        result: list = []
        for value in values:
            tmp = value
            if value < 0:
                tmp += 4096
            result.append(RenderNote.encodeBase64Core(int(tmp/64)) + RenderNote.encodeBase64Core(tmp&63))
        return result

    @staticmethod
    def encodeRunLength(values: list) -> str:
        '''
        | base64エンコードした文字列のリストを受け取り、ランレングス圧縮します。
        | 2文字一組とし、#num#はひとつ前の組の繰り返し回数を表します。
    
        >>> [AA, AB, AC] → AAABAC
        >>> [AA, AA, AA, AA] → AA#3#


        Parameters
        ----------
        values: list

        Returns
        -------
        result: str
        '''
        result: str = ""
        tmp: str = ""
        num: int = 0
        for value in values:
            if tmp != value:
                tmp = value
                if num != 0:
                    result += "#{}#".format(num)
                num = 0
                result += value
            else:
                num += 1
        if num != 0:
            result += "#{}#".format(num)
        return result


import os.path

from .Entry import *
from ..voicebank.prefixmap import PrefixMap
from ..voicebank.oto import Oto


class Note:
    '''
    ustやプラグインでのNoteを扱います。

    Attributes
    ----------
    num: NumberEntry
        ノート番号。#0000や#INSERT,#DELETEなどの形で与えられます。

    length: LengthEntry
        ノートのtick数。4分音符=480

    lyric: LyricEntry
        ノートの歌詞

    notenum: NoteNumEntry
        ノートの音高。C4=60

    tempo: TempoEntry
        ノートのbpm。


    pre: PreEntry
        ノートの先行発声


    atPre: AtPreEntry
        | UTAUのパラメータ自動調整適用後の先行発声値
        | 前の音が休符以外の場合

            >>> AtPre = pre / (pre -ove) * prev.msLength/2

        | 前の音が休符の場合

            >>> AtPre = pre / (pre -ove) * prev.msLength

    ove: OveEntry
        ノートのオーバーラップ値。


    atOve: AtOveEntry
        | UTAUのパラメータ自動調整適用後のオーバーラップ値
        | 前の音が休符以外の場合

            >>> AtOve = ove / (pre -ove) * prev.msLength/2

        | 前の音が休符の場合

            >>> AtOve = ove / (pre -ove) * prev.msLength

    stp: StpEntry
        ノートのstp値

    atStp: AtStpEntry
        | UTAUのパラメータ自動調整適用後のオーバーラップ値

            >>> atStp = Pre - atPre + stp

    atFileName: AtFileNameEntry
        このノートが参照しているファイルの音源ルートからの相対パス

    atAlias: AtAliasEntry
        このノートが参照している原音設定のエントリ

    velocity: VelocityEntry
        | このノートの子音速度。
        | 音源を再生するとき、固定範囲、先行発声、オーバーラップに以下の係数がかかる。

            >>> rate = 2 ** ((100-velocity)/100)

    intensity: IntensityEntry
        | このノートの音量。
        | resamplerのターゲット音量が、100で-6dB、200で0dB、0で-infとなる。

    modulation: ModulationEntry
        | このノートのモジュレーション
        | 0の時は音高通りの音程で、100の時は原音の音高のぶれを再現する。

    pitches: PitchesEntry
        mode1で扱うピッチ数列。

    pbStart: PBStartEntry
        ノートの頭から見てmode1のピッチのが何msから始まるか

    pbs: PBSEntry
        | mode2のピッチの開始点。以下のいずれかの書式で与えられる。

            >>> time;height
            >>> time,height
            >>> time

        | timeの単位はms、heightの単位はcent

    pby: PBYEntry
        mode2のピッチ制御点の音高列。制御点の数-2個(最初と最後を除く)与えられる。単位はcent

    pbw: PBWEntry
        mode2のピッチの制御点どおしの間隔を表す時間列。制御点の数-1個与えられる。単位はms

    pbm: PBMEntry
        | mode2のピッチの制御点どおしをどのように補完するかを表す値列。制御点の数-1個与えられる。
        | ""の場合、cos(-pi) → cos(0)の補完
        | "s"の場合、線形保管
        | "r"の場合、sin(0) → sin(pi/2)の補完
        | "j"の場合、cos(0) → cos(pi/2)の補完

    envelope: EnvelopeEntry
        | このノートの音量の変化。以下のいずれかの書式で与えられる。pはfloat(ms)、vはint

            >>> p1,p2,p3,v1,v2,v3,v4
            >>> p1,p2,p3,v1,v2,v3,v4,%,p4
            >>> p1,p2,p3,v1,v2,v3,v4,%,p4,p5,v5

        | p1はノート頭からのms
        | p2はp1からのms
        | p3はp4から前向きのms
        | p4はノート末尾から前向きのms
        | p5はp2からのms

    vibrato: VibratoEntry
        | このノートのビブラート。
        | lengthはノート長に対するビブラートがかかる範囲の割合
        | cycleはビブラートにおけるsin派1周にかかる時間(ms)
        | depthはビブラートのsin派の高さ(cent)
        | fadeInTimeはビブラートの波の大きさが最大になるまでの時間をビブラート全体の長さに対する割合で指定
        | fadeOutTimeはビブラートの波の大きさが0になるまでの時間をビブラート全体の長さに対する割合で指定
        | phaseは位相がsin派一周の何%ずれているか
        | heightは、0のとき波の中心が、-100のとき波の頂点が0に、100のとき波の底が0となるような割合

    label: LabelEntry
        このノートにつけられたラベル

    direct: DirectEntry
        Trueのとき、resamplerを経由せず直接wavtoolに値が送られる。

    region: RegionEntry
        「選択範囲に名前を付ける」でつけた名前の開始位置

    region_end: RegionEndEntry
        「選択範囲に名前を付ける」でつけた名前の終了位置

    flags: FlagsEntry
        このノートに適用されるのフラグ

    prev: Note, default None
        1つ前のノートへのポインタ

    next: note, default None
        1つ後ろのノートへのポインタ
    '''
    num: NumberEntry
    length: LengthEntry
    lyric: LyricEntry
    notenum: NoteNumEntry
    tempo: TempoEntry
    pre: PreEntry
    atPre: AtPreEntry
    ove: OveEntry
    atOve: AtOveEntry
    stp: StpEntry
    atStp: AtStpEntry
    atFileName: AtFileNameEntry
    atAlias: AtAliasEntry
    velocity: VelocityEntry
    intensity: IntensityEntry
    modulation: ModulationEntry
    pitches: PitchesEntry
    pbStart: PBStartEntry
    pbs: PBSEntry
    pby: PBYEntry
    pbm: PBMEntry
    pbw: PBWEntry
    envelope: EnvelopeEntry
    vibrato: VibratoEntry
    label: LabelEntry
    direct: DirectEntry
    region: RegionEntry
    region_end: RegionEndEntry
    flags: FlagsEntry
    prev = None
    next = None

    def __init__(self):
        self.num = NumberEntry()
        self.length = LengthEntry()
        self.lyric = LyricEntry()
        self.notenum = NoteNumEntry()
        self.tempo = TempoEntry()
        self.pre = PreEntry()
        self.atPre = AtPreEntry()
        self.ove = OveEntry()
        self.atOve = AtOveEntry()
        self.stp = StpEntry()
        self.atStp = AtStpEntry()
        self.atFileName = AtFileNameEntry()
        self.atAlias = AtAliasEntry()
        self.velocity = VelocityEntry()
        self.intensity = IntensityEntry()
        self.modulation = ModulationEntry()
        self.pitches = PitchesEntry()
        self.pbStart = PBStartEntry()
        self.pbs = PBSEntry()
        self.pby = PBYEntry()
        self.pbm = PBMEntry()
        self.pbw = PBWEntry()
        self.envelope = EnvelopeEntry()
        self.vibrato = VibratoEntry()
        self.label = LabelEntry()
        self.direct = DirectEntry()
        self.region = RegionEntry()
        self.region_end = RegionEndEntry()
        self.flags = FlagsEntry()
        self.autoren= bool = False
        self.prev = None
        self.next = None

    def apply_oto(self, oto: Oto, prefix: PrefixMap):
        '''
        | 原音設定値を読み込んで、pre,oveを更新します。
        | もし、パラメータが初期化されていない場合、atPre,atOve,atStp,atAlias,atFileNameも更新します。

        Parameters
        ----------
        oto: Oto
            原音設定ファイル

        prefix: prefixMap
            エイリアスの推定に使用します。

        Raises
        ------
        ValueError
            lyricもしくはnotenumが初期化されていない場合
        '''
        alias: str
        if not self.lyric.hasValue:
            raise ValueError("lyric is not initial")
        if not self.notenum.hasValue:
            raise ValueError("notenum is not initial")

        alias = self._init_alias(oto, prefix)
        self._apply_oto_to_pre(alias, oto)
        self._apply_oto_to_ove(alias, oto)
        self.autofit_atparam()

    def _init_alias(self, oto: Oto, prefix: PrefixMap) -> str:
        '''
        | 歌詞、音高、prefix.mapを参照してエイリアスを特定します。
        | 一致する原音設定レコードが見つからない場合、""を返します。
        | atAliasが値を持っていなかった場合、atAlias,atFileNameを更新します。

        Parameters
        ----------
        oto: Oto
            原音設定ファイル

        prefix: prefixMap
            エイリアスの推定に使用します。

        Returns
        -------
        alias: str, default ""
        '''

        lyric: str = self.lyric.value
        no_prefix: bool = False
        if "?" in lyric:
            lyryc = lyric.replace("?","")
            no_prefix = True
            
        if "!" in lyric:
            lyryc = lyric.replace("!","")
            self.autoren = True

        if self.atAlias.hasValue:
            return self.atAlias.value
        elif not no_prefix and oto.haskey(prefix[self.notenum.value].prefix + lyric + prefix[self.notenum.value].suffix):
            self.atAlias.value = prefix[self.notenum.value].prefix + lyric + prefix[self.notenum.value].suffix
            self.atFileName.value = os.path.join(oto[self.atAlias.value].otopath, oto[self.atAlias.value].filename)
            return prefix[self.notenum.value].prefix + lyric + prefix[self.notenum.value].suffix
        elif oto.haskey(lyric):
            self.atAlias.value = lyric
            self.atFileName.value = os.path.join(oto[lyric].otopath, oto[lyric].filename)
            return lyric
        else:
            return ""

    def _apply_oto_to_pre(self, alias: str, oto: Oto):
        '''
        | 原音設定値を読み込んで、preを更新します。
        | もしalias=""の場合、0で更新します。

        Parameters
        ----------
        alias: str
            otoのkeyになるエイリアス
        oto: Oto
            原音設定ファイル
        '''
        if self.pre.hasValue:
            return
        if alias == "":
            self.pre.value = 0
        else:
            self.pre.value = oto[alias].pre
        self.pre.hasValue = False

    def _apply_oto_to_ove(self, alias: str, oto: Oto):
        '''
        | 原音設定値を読み込んで、oveを更新します。
        | もしalias=""の場合、0で更新します。

        Parameters
        ----------
        alias: str
            otoのkeyになるエイリアス
        oto: Oto
            原音設定ファイル
        '''
        if self.ove.hasValue:
            return
        if alias == "":
            self.ove.value = 0
        else:
            self.ove.value = oto[alias].ove
        self.ove.hasValue = False

    def autofit_atparam(self):
        '''
        pre,ove,stp,velocity,prev.length,prev.tempoを勘案して、atpre,atove,atstpを更新します。

        Raises
        ------
        ValueError
            prev.lengthがNone出ないにもかかわらず、lyric,length,tempoの値が与えられていないとき。
        '''
        if self.prev is None:
            # 前のノートが存在しない場合、自動調整は不要。
            self.atPre.value = self.pre.value
            self.atOve.value = self.ove.value
            self.atStp.value = self.stp.value
            return

        if not self.prev.lyric.hasValue:
            raise ValueError("prev lyric is not initial")

        prevMsLength: int = self.prev.msLength
        realPre: float = self.pre.value * self.velocity.rate
        realOve: float = self.ove.value * self.velocity.rate
        realStp: float = self.stp.value * self.velocity.rate
        if self.prev.lyric.value != "R":
            prevMsLength /= 2

        if prevMsLength < (realPre - realOve):
            self.atPre.value = realPre / (realPre - realOve) * prevMsLength
            self.atOve.value = realOve / (realPre - realOve) * prevMsLength
            self.atStp.value = realPre - self.atPre.value + realStp
        else:
            self.atPre.value = realPre
            self.atOve.value = realOve
            self.atStp.value = realStp

    @property
    def msLength(self) -> float:
        '''
        tempoとlengthからmsを計算して返します。

        Returns
        -------
        msLength: int

        Raises
        ------
        ValueError
            tempoもしくはlengthが初期化されていないとき。
        '''
        if not self.length.hasValue:
            raise ValueError("length is not initial")

        #if not self.tempo.hasValue:
        #    raise ValueError("tempo is not initial")

        return 60 / self.tempo.value * self.length.value / 480 * 1000



'''
windows版UTAUの情報を取得します。
'''


import platform
if platform.platform().startswith("Windows"):
    import winreg
import os
import os.path


def is_utau_installed() -> bool:
    '''
    UTAUがインストールされているか確認します。

    Returns
    -------
    is_utau_installed: bool

        | レジストリのHKEY_CLASSES_ROOT\\\\UTAUSequenceTextが存在すれば、インストール済みとみなしTrueを返します。
        | windows以外の環境ではFalseを返します。
    '''
    if not platform.platform().startswith("Windows"):
        return False

    subkeys: list = []
    with winreg.OpenKey(winreg.HKEY_CLASSES_ROOT, "") as key:
        subKeyNum: int = winreg.QueryInfoKey(key)[0]
        for i in range(subKeyNum):
            subkeys.append(winreg.EnumKey(key, i))

    return "UTAUSequenceText" in subkeys


def get_utau_root() -> str:
    '''
    UTAU.exeの場所を返します。

    Returns
    -------
    dir_name: str

        | UTAU.exeの場所
        | レジストリのHKEY_CLASSES_ROOT\\\\UTAUSequenceText\\\\shell\\\\open\\\\commandから取得します。
        | UTAUがインストールされていない場合、""を返します。
        | windows以外の環境では""を返します。

    '''

    if not platform.platform().startswith("Windows"):
        return ""
    if not is_utau_installed():
        return ""
    reg_value: str
    with winreg.OpenKey(winreg.HKEY_CLASSES_ROOT, "UTAUSequenceText\\shell\\open\\command") as key:
        reg_value = winreg.EnumValue(key, 0)
    return os.path.dirname(" ".join(reg_value[1].split(" ")[:-1]).replace("\"", ""))


def get_utau_settings(utau_root: str="") -> dict:
    '''
    | UTAUのsetting.iniを読み込みます。
    | 読み込む優先順位は下記のとおりです。

    1. (UTAUが program files(x86)にインストールされている場合) %LOCALAPPDATA%\\\\VirtualStore\\\\program files(x86)\\\\UTAU\\\\setting.ini
    2. (ほかのフォルダにインストールされているか、上記のフォルダになければ) {UTAU_ROOT}\\\\setting.ini

    Parameters
    ----------
    utau_root: str, default ""
        UTAUのパス
        初期値の場合、*get_utau_root* を呼び出します。

    Returns
    -------
    utau_settings: dict
        UTAUのsetting.iniの内容を読み込んだ辞書ファイル
        見つからなければ空の辞書を返します。

    '''
    if utau_root == "":
        utau_root = get_utau_root()

    if utau_root == "":
        return {}

    settings_path: str = os.path.join(utau_root, "setting.ini")
    read_datas: list
    utau_settings: dict = {}
    if "program files(x86)" in utau_root:
        path_tail: str = utau_root.split("program files(x86)")[1].replace("\\", "")
        if os.path.isfile(os.path.join(os.environ["localappdata"], "VirtualStore", "program files(x86)", path_tail, "setting.ini")):
            settings_path = os.path.join(os.environ["localappdata"], "VirtualStore", "program files(x86)", path_tail, "setting.ini")

    with open(settings_path, "r") as fr:
        read_datas = fr.read().replace("\r", "").split("\n")

    for data in read_datas:
        if "=" not in data:
            continue
        tmp = data.split("=")
        utau_settings[tmp[0]] = tmp[1]

    return utau_settings



'''frq

frq関係のデータを扱います。

'''

import os
import os.path
import struct
import wave
from typing import Tuple


import numpy as np
import pyworld as pw


class Frq:
    '''
    frqファイルを扱います。

    Attributes
    ----------
    frqpath: str
        | frqファイルのパス

    framerate: int, default = 44100
        wavのサンプリング周波数

    f0: np.ndarray of np.float64
        f0データの1次元配列

    f0_avg: float
        平均キー

    amp: np.ndarray of np.float64
        音量データの1次元配列

    t: np.ndarray
        時間配列。原則として256sample毎に与えられる。
    '''

    frqpath: str
    framerate: int = 44100
    f0: np.ndarray
    amp: np.ndarray
    t: np.ndarray

    def __init__(self, frqpath: str = "", framerate: int = 44100):
        '''
        Parameters
        ----------
        frqpath: str, default = ""
            frqのパス。

        framerate: int, default 44100
            wavのサンプリング周波数

        Raises
        ------
        FileNotFoundError
            指定したfrqファイルが存在しないとき
        '''
        self.frqpath = frqpath
        self.framerate = framerate
        if frqpath != "":
            self.load(frqpath, framerate)

    def load(self, frqpath: str, framerate: int):
        '''
        frqpathで指定したファイルを開き、self.f0, self.f0_avg, self.amp, self.tを更新します。

        Parameters
        ----------
        frqpath: str
            frqのパス。

        framerate: int
            wavのサンプリング周波数

        Raises
        ------
        FileNotFoundError
            指定したfrqファイルが存在しないとき
        '''
        if not os.path.isfile(frqpath):
            raise FileNotFoundError("{} is not found.".format(frqpath))
        with open(frqpath, "rb") as fr:
            bytes_data = fr.read()
        self.f0_avg = struct.unpack("<d", bytes_data[12:20])[0]
        data: np.ndarray = np.frombuffer(bytes_data[40:], dtype="float64")
        self.f0 = data[::2]
        self.amp = data[1::2]
        frq_span = 1 / framerate * 256
        self.t = np.arange(0, frq_span * (self.f0.shape[0] + 1), frq_span)[:self.f0.shape[0]]

    def save(self):
        '''
        self.frqpathで指定したパスにfrqファイルを保存します。
        '''
        with open(self.frqpath, "wb") as fw:
            fw.write(b"FREQ0003")
            fw.write((256).to_bytes(4, "little"))
            fw.write(np.array(self.f0_avg, dtype=np.float64).tobytes())
            fw.write(b"PyUtauCli       ")
            fw.write(self.f0.shape[0].to_bytes(4, "little"))
            fw.write(np.concatenate([[self.f0], [self.amp]]).T.tobytes())

    def make(self, wavpath: str):
        '''
        wavpathで指定したwavファイルを解析し、self.frqpath, self.framerate, self.f0, self.f0_avg, self.amp, self.tを更新します。

        Parameters
        ----------
        wavpath: str
            wavファイルの絶対パス

        Raises
        ------
        FileNotFoundError
            wavpathのwavファイルが見つからなかったとき
        wave.Error
            input_pathで指定したファイルがwavではなかったとき
        '''
        datas, self.framerate = self._wave_load(wavpath)
        self.frqpath = ".".join(wavpath.split(".")[:-1]) + "_wav.frq"
        self.f0, self.t = pw.harvest(datas, self.framerate, frame_period=1000 / self.framerate * 256)
        self.f0 = pw.stonemask(datas, self.f0, self.t, self.framerate)
        self.amp = np.zeros_like(self.f0)
        for i in range(self.amp.shape[0]):
            self.amp[i] = np.average(np.abs(datas[i * 256:(i + 1) * 256]))

    def _wave_load(self, input_path) -> Tuple[np.ndarray, int]:
        '''
        wavファイルを読み込み、データとサンプルレートを返します。

        Parameters
        ----------
        input_path: str
            原音のファイル名

        Returns
        -------
        data: np.ndarray or np.float64
            指定された区間のwaveのデータ。1次元
        framerate: int
            wavのサンプリング周波数

        Raises
        ------
        FileNotFoundError
            input_pathにwaveファイルがなかったとき
        wave.Error
            input_pathで指定したファイルがwavではなかったとき
        '''

        if not os.path.exists(input_path):
            raise FileNotFoundError("{} not found.".format(input_path))
        with wave.open(input_path, "rb") as wr:
            channels: int = wr.getnchannels()
            framerate: int = wr.getframerate()
            sampwidth: int = wr.getsampwidth()
            nframes: int = wr.getnframes()  # フレーム数
            bytes_data: byte = wr.readframes(nframes)

        if sampwidth == 1:
            data = np.frombuffer(bytes_data, dtype="int8")
        elif sampwidth == 2:
            data = np.frombuffer(bytes_data, dtype="int16")
        elif sampwidth == 3:
            data = np.zeros(int((end_frame - offset_frame)), dtype="int32")
            for i in range(int(len(bytes_data) / sampwidth)):
                data[i] = int.from_bytes(bytes_data[i * sampwidth:(i + 1) * sampwidth], "little", signed=True)
        elif sampwidth == 4:
            data = np.frombuffer(bytes_data, dtype="int32")
        if channels == 2:
            data = data[::2]

        return data / 2**(sampwidth * 8 - 1), framerate



import os.path

from ..common import convert_notenum as convert_notenum


class MapRecord:
    '''PrefixMapの各行のデータを扱います。

    Attributes
    ----------
    key: str

        | 音名。C4,D#3,Ab5など
        | この値は、loadしたprefix.mapをsaveするときに活用します。
        | 参照には活用しません。

    prefix: str
    suffix: str
    '''

    key: str
    prefix: str
    suffix: str

    def __init__(self, record: str):
        '''
        Parameters
        ----------
        record: str
            key,prefix,suffixの順でタブで区切った文字列

        Raises
        ------
        ValueError
            レコードのフォーマットに則らない文字列が渡されたとき
        '''
        try:
            self.key, self.prefix, self.suffix = record.split("\t")
        except:
            raise ValueError("can't set prefix map value.{} is bad format.".format(record))


class PrefixMap:
    '''prefix.mapのデータを扱います。
    '''
    _key: list = [i for i in range(24, 108)]
    _values: dict = {}

    def __init__(self, dirpath: str = ""):
        '''
        Parameters
        ----------
        dirpath: str, default ""
            音源のルートディレクトリのパス。引数が与えられている場合self.load(dirpath)を実行する。

        Raises
        ------
        FileNotFoundError
            load実行時prefix.mapが見つからなかった場合
        UnicodeDecodeError
            load実行時ファイルがcp932でもutf-8でもなかった場合
        '''
        for i in self._key:
            self._values[i] = MapRecord(convert_notenum.toStr(i) + "\t\t")
        if dirpath != "":
            self.load(dirpath)

    def __getitem__(self, key) -> MapRecord:
        if key in self._values:
            return self._values[key]
        return self._values[convert_notenum.toInt(key)]

    def load(self, dirpath: str, filename: str = "prefix.map"):
        '''
        *dirpath\\\\prefix.map* を読み込んでself._valueを更新する。

        Parameters
        ----------
        dirpath: str
            音源のルートディレクトリのパス

        filename: str, default "prefix.map"
            ファイル名。UTAU本体の仕様では指定不要ですが、複数のprefix.mapを読み込む仕様を想定

        Raises
        ------
        FileNotFoundError
            prefix.mapが見つからなかった場合
        UnicodeDecodeError
            ファイルがcp932でもutf-8でもなかった場合
        '''
        filepath: str = os.path.join(dirpath, filename)
        lines: list
        if not os.path.isfile(filepath):
            raise FileNotFoundError("{} is not found.".format(filepath))

        try:
            with open(filepath, "r", encoding="cp932") as fr:
                lines = fr.read().replace("\r", "").split("\n")
        except:
            try:
                with open(filepath, "r", encoding="utf-8") as fr:
                    lines = fr.read().replace("\r", "").split("\n")
            except UnicodeDecodeError as e:
                e.reason = "can't read prefix.map. because required character encoding is utf-8 or cp932"
                raise e

        for line in lines:
            if line == "":
                continue
            if "\t" not in line:
                continue
            self._values[convert_notenum.toInt(line.split("\t")[0])] = MapRecord(line)

    def save(self, dirpath: str, filename: str = "prefix.map", encoding: str = "cp932"):
        '''
        *dirpath\\\\prefix.map* を *encoding* で保存する。

        Parameters
        ----------
        dirpath: str
            音源のルートディレクトリのパス

        filename: str, default "prefix.map"
            ファイル名。UTAU本体の仕様では指定不要ですが、複数のprefix.mapを読み込む仕様を想定

        encoding: str, default "cp932"

            | 保存する際の文字コード
            | UTAU本体のGUIを考えるとcp932が無難

        Raises
        ------
        OSError
            character.txtに書き込み権限がないとき
        UnicodeDecodeError
            パラメータをcp932に変換できなかったとき
        '''
        filepath: str = os.path.join(dirpath, filename)
        with open(filepath, "w", encoding=encoding) as fw:
            for i in range(len(self._values)):
                fw.write(self._values[107 - i].key + "\t" + self._values[107 - i].prefix + "\t" + self._values[107 - i].suffix + "\r\n")



'''Settins
各種定数や初期値の設定

Attributes
----------
UTAU_IS_INSTALLED: bool

    | UTAUがインストール済みか否か
    | windows以外の環境では自動的にFalseになる。

UTAU_ROOT: str, default ""

    | UTAUがインストールされたフォルダのパス

UTAU_SETTINGS: dict, default {}

    | UTAUのsetting.iniの内容を読み込んだ辞書ファイル

VOICE_ROOT: list of str

    | 音源フォルダのルート
    | windows環境では、UTAU.exeと同じ階層のsetting.iniに記載あり

'''

import platform
import os
import os.path

from . import win_utau

UTAU_IS_INSTALLED: bool = win_utau.is_utau_installed()
UTAU_SETTINGS: dict = {}
VOICE_ROOT: list = []
if UTAU_IS_INSTALLED:
    UTAU_ROOT: str = win_utau.get_utau_root()
    UTAU_SETTINGS = win_utau.get_utau_settings()
    if "VoiceRoot" in UTAU_SETTINGS:
        VOICE_ROOT.append(UTAU_SETTINGS["VoiceRoot"])
    VOICE_ROOT.append(os.path.join(os.environ["appdata"], "UTAU", "voice"))
    VOICE_ROOT.append(os.path.join(UTAU_ROOT, "voice"))
else:
    UTAU_ROOT: str = ""
    
DEFAULT_ENV: str = UTAU_SETTINGS.get("DefaultEnv", "0 10 35 0 100 100 0")


import os.path


class Character:
    '''Character.txtを扱います

    Attributes
    ----------
    name: str, default ""
        VBの名前、与えられないときはフォルダ名

    image: str, default ""
        アイコン画像のパス

    sample: str, default ""
        サンプル音声のパス

    author: str, default ""
        管理人名

    web: str, default ""
        webサイトurl

    version: str, default ""
        バージョン

    dirpath: str, default ""
        音源のルートのパス
    '''
    _name: str = ""
    image: str = ""
    sample: str = ""
    author: str = ""
    web: str = ""
    version: str = ""
    dirpath: str = ""

    @property
    def name(self) -> str:
        if self._name != "":
            return self._name
        elif self.dirpath == "":
            return ""
        else:
            return os.path.split(self.dirpath)[1]

    @name.setter
    def name(self, value):
        self._name = value

    def __init__(self, dirpath: str = ""):
        '''
        Parameters
        ----------
        dirpath: str, default ""
            音源のルートディレクトリのパス。引数が与えられている場合self.load(dirpath)を実行する。

        Raises
        ------
        FileNotFoundError
            read実行時character.txtが見つからなかった場合
        UnicodeDecodeError
            read実行時ファイルがcp932でも環境の文字コードでもなかった場合
        '''
        if dirpath != "":
            self.load(dirpath)
            
    def load(self, dirpath: str):
        '''
        *dirpath\\\\character.txt* を読み込んで各パラメータを更新する。

        Parameters
        ----------
        dirpath
            音源のルートディレクトリのパス

        Raises
        ------
        FileNotFoundError
            character.txtが見つからなかった場合
        UnicodeDecodeError
            ファイルがcp932でも環境の文字コードでもなかった場合
        '''
        filepath: str = os.path.join(dirpath, "character.txt")
        lines: list
        self.dirpath = dirpath
        if not os.path.isfile(filepath):
            raise FileNotFoundError("{} is not found.".format(filepath))

        try:
            with open(filepath, "r") as fr:
                lines = fr.read().replace("\r", "").split("\n")
        except:
            try:
                with open(filepath, "r", encoding="cp932") as fr:
                    lines = fr.read().replace("\r", "").split("\n")
            except UnicodeDecodeError as e:
                e.reason = "can't read character.txt. because required character encoding is system default or cp932"
                raise e
        key: str
        value: str
        for line in lines:
            if line == "":
                continue
            if "=" not in line and ":" not in line:
                continue
            if "=" in line:
                key, value = line.split("=")
            else:
                key = line.split(":")[0]
                value = ":".join(line.split(":")[1:])

            if key.lower() == "name":
                self._name = value
            elif key.lower() == "image":
                self.image = value
            elif key.lower() == "sample":
                self.sample = value
            elif key.lower() == "author":
                self.author = value
            elif key.lower() == "web":
                self.web = value
            elif key.lower() == "version":
                self.version = value

    def save(self, encoding: str="cp932"):
        '''
        *self.dirpath\\\\character.txt* を *encoding* で保存する。

        Parameters
        ----------
        encoding: str, default "cp932"
            
            | 保存する際の文字コード
            | UTAU本体のGUIを考えるとcp932が無難

        Raises
        ------
        OSError
            character.txtに書き込み権限がないとき
        UnicodeDecodeError
            パラメータをcp932に変換できなかったとき
        '''
        filepath: str = os.path.join(self.dirpath, "character.txt")
        with open(filepath, "w", encoding=encoding) as fw:
            if self._name != "":
                fw.write("name=" + self._name + "\r\n")
            if self.image != "":
                fw.write("image=" + self.image + "\r\n")
            if self.sample != "":
                fw.write("sample=" + self.sample + "\r\n")
            if self.author != "":
                fw.write("Author=" + self.author + "\r\n")
            if self.web != "":
                fw.write("Web=" + self.web + "\r\n")
            if self.version != "":
                fw.write("Version=" + self.version + "\r\n")



import os
import os.path

from .Ust import Ust

class UtauPlugin(Ust):
    '''
    | UTAUのプラグイン用一時ファイルを扱います。
    | ほぼ、Ustと共通の仕様ですが、主に書き出しに関する仕様が異なります。
    '''
    def save(self, filepath: str = "", encoding: str = "cp932"):
        '''
        | self.filepathもしくはfilepathにファイルを保存します。
        | windows版UTAUとの互換性を優先してcp932を優先します。

        Parameters
        ----------
        filepath: str, default ""
        encoding: str, default "cp932"
        '''
        if filepath != "":
            self.filepath = filepath
        if os.path.split(self.filepath)[0] != "":
            os.makedirs(os.path.split(self.filepath)[0], exist_ok=True)
        self.logger.info("saving utau plugin temp to:{}".format(self.filepath))
        with open(self.filepath, "w", encoding=encoding) as fw:
            for note in self.notes:
                fw.write("[{}]\n".format(str(note.num)))
                if note.num.value == "#DELETE":
                    continue
                if note.length.hasValue and note.length.isUpdate:
                    fw.write("Length={}\n".format(str(note.length)))
                if note.lyric.hasValue and note.lyric.isUpdate:
                    fw.write("Lyric={}\n".format(str(note.lyric)))
                if note.notenum.hasValue and note.notenum.isUpdate:
                    fw.write("NoteNum={}\n".format(str(note.notenum.value)))
                if note.tempo.hasValue and note.tempo.isUpdate:
                    fw.write("Tempo={}\n".format(str(note.tempo)))
                if note.pre.hasValue and note.pre.isUpdate:
                    fw.write("PreUtterance={}\n".format(str(note.pre)))
                if note.ove.hasValue and note.ove.isUpdate:
                    fw.write("VoiceOverlap={}\n".format(str(note.ove)))
                if note.stp.hasValue and note.stp.isUpdate:
                    fw.write("StartPoint={}\n".format(str(note.stp)))
                if note.velocity.hasValue and note.velocity.isUpdate:
                    fw.write("Velocity={}\n".format(str(note.velocity)))
                if note.intensity.hasValue and note.intensity.isUpdate:
                    fw.write("Intensity={}\n".format(str(note.intensity)))
                if note.modulation.hasValue and note.modulation.isUpdate:
                    fw.write("Modulation={}\n".format(str(note.modulation)))
                if note.pitches.hasValue and note.pitches.isUpdate:
                    fw.write("Pitches={}\n".format(str(note.pitches)))
                if note.pbStart.hasValue and note.pbStart.isUpdate:
                    fw.write("PBStart={}\n".format(str(note.pbStart)))
                if note.pbs.hasValue and note.pbs.isUpdate:
                    fw.write("PBS={}\n".format(str(note.pbs)))
                if note.pby.hasValue and note.pby.isUpdate:
                    fw.write("PBY={}\n".format(str(note.pby)))
                if note.pbm.hasValue and note.pbm.isUpdate:
                    fw.write("PBM={}\n".format(str(note.pbm)))
                if note.pbw.hasValue and note.pbw.isUpdate:
                    fw.write("PBW={}\n".format(str(note.pbw)))
                if note.flags.hasValue and note.flags.isUpdate:
                    fw.write("Flags={}\n".format(str(note.flags)))
                if note.vibrato.hasValue and note.vibrato.isUpdate:
                    fw.write("VBR={}\n".format(str(note.vibrato)))
                if note.envelope.hasValue and note.envelope.isUpdate:
                    fw.write("Envelope={}\n".format(str(note.envelope)))
                if note.label.hasValue and note.label.isUpdate:
                    fw.write("Label={}\n".format(str(note.label)))
                if note.direct.hasValue and note.direct.isUpdate:
                    fw.write("$direct={}\n".format(str(note.direct)))
                if note.region.hasValue and note.region.isUpdate:
                    fw.write("$region={}\n".format(str(note.region)))
                if note.region_end.hasValue and note.region_end.isUpdate:
                    fw.write("$region_end={}\n".format(str(note.region_end)))
        self.logger.info("saving utau plugin temp to:{} complete".format(self.filepath))


from .EntryBase import EntryBase, IntEntry, StringEntry, FloatEntry, BoolEntry, ListEntry
from ..common import convert_notenum as convert_notenum


class NumberEntry(StringEntry):
    pass


class LengthEntry(IntEntry):
    _value = 480


class LyricEntry(StringEntry):
    pass


class NoteNumEntry(IntEntry):
    _value = 60

    def set_from_str(self, value: str):
        self.value = convert_notenum.toInt(value)
        self._set_update()
        self._hasValue = True

    def get_tone_name(self) -> str:
        return convert_notenum.toStr(self.value)


class TempoEntry(FloatEntry):
    _value = 120
    point = 2

    @property
    def hasValue(self) -> bool:
        return self._hasValue

    @hasValue.setter
    def hasValue(self, value: bool):
        self._hasValue = value


class PreEntry(FloatEntry):
    @property
    def hasValue(self) -> bool:
        return self._hasValue

    @hasValue.setter
    def hasValue(self, value: bool):
        self._hasValue = value


class AtPreEntry(FloatEntry):
    pass


class OveEntry(FloatEntry):
    @property
    def hasValue(self) -> bool:
        return self._hasValue

    @hasValue.setter
    def hasValue(self, value: bool):
        self._hasValue = value


class AtOveEntry(FloatEntry):
    pass


class StpEntry(FloatEntry):
    pass


class AtStpEntry(FloatEntry):
    pass


class AtFileNameEntry(StringEntry):
    pass


class AtAliasEntry(StringEntry):
    pass


class VelocityEntry(IntEntry):
    _value = 100

    @property
    def rate(self) -> float:
        return 2 ** ((100 - self._value) / 100)


class IntensityEntry(IntEntry):
    _value = 100


class ModulationEntry(IntEntry):
    _value = 100


class PitchesEntry(ListEntry):
    def _check_value(self, value):
        try:
            return int(value)
        except:
            raise ValueError("{} is not int".format(value))


class PBStartEntry(FloatEntry):
    pass


class PBSEntry(EntryBase):
    _time: float = 0
    _height: float = 0
    _value: str

    @property
    def time(self) -> float:
        return self._time

    @property
    def height(self) -> float:
        return self._height

    @property
    def value(self) -> str:
        return self._value

    @time.setter
    def time(self, value: float):
        try:
            self._time = float(value)
            self._value = "{:.3f};{:.3f}".format(self._time, self._height)
            self._set_update()
            self._hasValue = True
        except:
            raise ValueError("{} is not float".format(value))

    @height.setter
    def height(self, value: float):
        try:
            self._height = float(value)
            self._value = "{:.3f};{:.3f}".format(self._time, self._height)
            self._set_update()
            self._hasValue = True
        except:
            raise ValueError("{} is not float".format(value))

    @value.setter
    def value(self, value: str):
        value = value.replace(",", ";")
        if ";" in value:
            values: list = value.split(";")
            try:
                self._time = float(values[0])
            except:
                raise ValueError("{} is not float".format(values[0]))
            try:
                self._height = float(values[1])
            except:
                raise ValueError("{} is not float".format(values[1]))
            self._value = value
            self._set_update()
            self._hasValue = True
        else:
            try:
                self._time = float(value)
                self._height = 0
            except:
                raise ValueError("{} is not float".format(value))
            self._value = value
            self._set_update()
            self._hasValue = True

    def init(self, value: str):
        value = value.replace(",", ";")
        if ";" in value:
            values: list = value.split(";")
            try:
                self._time = float(values[0])
            except:
                raise ValueError("{} is not float".format(values[0]))
            try:
                self._height = float(values[1])
            except:
                raise ValueError("{} is not float".format(values[1]))
            self._value = value
            self._hasValue = True
        else:
            try:
                self._time = float(value)
                self._height = 0
            except:
                raise ValueError("{} is not float".format(value))
            self._value = value
            self._hasValue = True

    def __str__(self) -> str:
        if self._height == 0:
            return "{:.3f}".format(self._time)
        else:
            return self._value


class PBYEntry(ListEntry):
    def _check_value(self, value):
        try:
            if value == " ":
                value = 0
            if value == "":
                value = 0
            return float(value)
        except:
            raise ValueError("{} is not float".format(value))


class PBWEntry(ListEntry):
    def _check_value(self, value):
        try:
            return float(value)
        except:
            raise ValueError("{} is not float".format(value))


class PBMEntry(ListEntry):
    def _check_value(self, value):
        if value in ["", "s", "r", "j"]:
            return str(value)
        else:
            raise ValueError("{} is not '',s,r,j".format(value))


class EnvelopeEntry(EntryBase):
    _value: str
    _p: list
    _v: list
    separater: str = ","

    @property
    def value(self) -> str:
        if len(self._p) == 3:
            return "{:.2f},{:.2f},{:.2f},{},{},{},{}".format(self._p[0], self._p[1], self._p[2], self._v[0], self._v[1], self._v[2], self._v[3])
        elif len(self._p) == 4:
            return "{:.2f},{:.2f},{:.2f},{},{},{},{},%,{:.2f}".format(self._p[0], self._p[1], self._p[2], self._v[0], self._v[1], self._v[2], self._v[3], self._p[3])
        elif len(self._p) == 5:
            return "{:.2f},{:.2f},{:.2f},{},{},{},{},%,{:.2f},{:.2f},{}".format(self._p[0], self._p[1], self._p[2], self._v[0], self._v[1], self._v[2], self._v[3], self._p[3], self._p[4], self._v[4])

        return self._value

    @property
    def p(self) -> list:
        return self._p[:]

    @property
    def v(self) -> list:
        return self._v[:]

    @value.setter
    def value(self, value: str):
        self._set_value(value)
        self._set_update()
        self._hasValue = True

    def _set_value(self, value):
        tmp: list = value.split(self.separater)
        self._p = []
        self._v = []
        try:
            for i in range(len(tmp)):
                if i in [0, 1, 2, 8, 9]:
                    self._p.append(float(tmp[i]))
                elif i in [3, 4, 5, 6, 10]:
                    self._v.append(int(tmp[i]))
        except:
            raise ValueError("{} is not envelope pattern".format(value))
        self._value = value

    def init(self, value: list):
        self._set_value(value)
        self._hasValue = True

    def __str__(self) -> str:
        return self.value

    def set_p(self, pos: int, value: float):
        self._p[pos] = float(value)
        self._set_update()

    def set_v(self, pos: int, value: int):
        self._v[pos] = int(value)
        self._set_update()


class VibratoEntry(EntryBase):
    _length: float
    _cycle: float
    _depth: float
    _fadeInTime: float
    _fadeOutTime: float
    _phase: float
    _height: float
    _amp: float
    _value: str
    separater: str = ","

    @property
    def length(self) -> float:
        return self._length

    @property
    def cycle(self) -> float:
        return self._cycle

    @property
    def depth(self) -> float:
        return self._depth

    @property
    def fadeInTime(self) -> float:
        return self._fadeInTime

    @property
    def fadeOutTime(self) -> float:
        return self._fadeOutTime

    @property
    def phase(self) -> float:
        return self._phase

    @property
    def height(self) -> float:
        return self._height

    @property
    def amp(self) -> float:
        return self._amp

    @property
    def value(self) -> str:
        return "{:.2f},{:.2f},{:.2f},{:.2f},{:.2f},{:.2f},{:.2f},{:.2f}".format(self._length,
                                                                         self._cycle,
                                                                         self._depth,
                                                                         self._fadeInTime,
                                                                         self._fadeOutTime,
                                                                         self._phase,
                                                                         self._height,
                                                                         self._amp)

    @length.setter
    def length(self, value: float):
        try:
            self._length = float(value)
        except:
            raise ValueError("{} is not float".format(value))
        self._set_update()

    @cycle.setter
    def cycle(self, value: float):
        try:
            self._cycle = float(value)
        except:
            raise ValueError("{} is not float".format(value))
        self._set_update()

    @depth.setter
    def depth(self, value: float):
        try:
            self._depth = float(value)
        except:
            raise ValueError("{} is not float".format(value))
        self._set_update()

    @fadeInTime.setter
    def fadeInTime(self, value: float):
        try:
            self._fadeInTime = float(value)
        except:
            raise ValueError("{} is not float".format(value))
        self._set_update()

    @fadeOutTime.setter
    def fadeOutTime(self, value: float):
        try:
            self._fadeOutTime = float(value)
        except:
            raise ValueError("{} is not float".format(value))
        self._set_update()

    @phase.setter
    def phase(self, value: float):
        try:
            self._phase = float(value)
        except:
            raise ValueError("{} is not float".format(value))
        self._set_update()

    @height.setter
    def height(self, value: float):
        try:
            self._height = float(value)
        except:
            raise ValueError("{} is not float".format(value))
        self._set_update()
        
    @amp.setter
    def amp(self, value: float):
        try:
            self._amp = float(value)
        except:
            raise ValueError("{} is not float".format(value))
        self._set_update()

    @value.setter
    def value(self, value: str):
        values: list = value.split(self.separater)
        self.length = values[0]
        self.cycle = values[1]
        self.depth = values[2]
        self.fadeInTime = values[3]
        self.fadeOutTime = values[4]
        self.phase = values[5]
        self.height = values[6]
        self.amp = values[7]
        self._hasValue = True

    def init(self, value: str):
        self.value = value
        self._isUpdate = False

    def __str__(self) -> str:
        return self.value


class LabelEntry(StringEntry):
    pass


class DirectEntry(BoolEntry):
    pass


class RegionEntry(StringEntry):
    pass


class RegionEndEntry(StringEntry):
    pass


class FlagsEntry(StringEntry):
    @property
    def hasValue(self) -> bool:
        return self._hasValue

    @hasValue.setter
    def hasValue(self, value: bool):
        self._hasValue = value



import os.path
import wave


class OtoRecord:
    '''oto.iniの各行のデータを扱います。

    Attributes
    ----------
    otopath: str
        oto.iniのパス。voicedirからの相対パスで与えられる
    filename: str
        wavファイルのファイル名
    alias: str
        エイリアス
    offset: float
        オフセット(setparamでいうところの左ブランク)
    pre: float
        先行発声
    ove: float
        オーバーラップ
    consonant: float
        子音部(setparamでいうところの固定範囲)
    blank: float
        | ブランク(setparamでいうところの右ブランク)
        | 正の値のときはwav末尾からのmsを表す。
        | 負の値のときはオフセットからのmsの絶対値に-1をかけたものを表す。
    '''

    otopath: str
    filename: str
    alias: str
    offset: float
    pre: float
    ove: float
    consonant: float
    blank: float

    def __init__(self, otopath: str, filename: str, alias: str, offset: float, pre: float, ove: float, consonant: float, blank: float):
        '''oto.iniの各行のデータを扱います。

        Parameters
        ----------
        otopath: str
            oto.iniのパス。voicedirからの相対パスで与えられる
        filename: str
            wavファイルのファイル名
        alias: str
            | エイリアス
            | 空の値が与えられた場合、os.path.join(otopath, filename)から拡張子を除いたもので初期化する。
        offset: float
            オフセット(setparamでいうところの左ブランク)
        pre: float
            先行発声
        ove: float
            オーバーラップ
        consonant: float
            子音部(setparamでいうところの固定範囲)
        blank: float
            | ブランク(setparamでいうところの右ブランク)
            | 正の値のときはwav末尾からのmsを表す。
            | 負の値のときはオフセットからのmsの絶対値に-1をかけたものを表す。
        '''
        self.otopath = otopath
        self.filename = filename
        if alias != "":
            self.alias = alias
        else:
            self.alias = ".".join(os.path.join(otopath, filename).split(".")[:-1])
        self.offset = float(offset)
        self.pre = float(pre)
        self.ove = float(ove)
        self.consonant = float(consonant)
        self.blank = float(blank)

    def invert_blank(self, dirpath: str):
        '''
        | self.blank値の正負を反転します。
        | blankは正の数の時、ノート末尾からの長さms
        | 負の数の時、その絶対値がoffsetからの長さmsを表します。

        Parameters
        ----------
        dirpath: str
            音源のルートディレクトリのパス。

        Raises
        ------
        FileNotFoundError
            os.path.join(dirpath, self.otopath, self.filename)のwavファイルが見つからなかったとき
        wave.Error
            os.path.join(dirpath, self.otopath, self.filename)がwavファイルではなかったとき
        '''
        if not os.path.isfile(os.path.join(dirpath, self.otopath, self.filename)):
            raise FileNotFoundError("{} is not found.".format(os.path.join(dirpath, self.otopath, self.filename)))
        with wave.open(os.path.join(dirpath, self.otopath, self.filename), "rb") as wr:
            wav_length = wr.getnframes() / wr.getframerate() * 1000
        if self.blank >= 0:
            self.blank = self.offset - (wav_length - self.blank)
        else:
            self.blank = wav_length - (self.offset - self.blank)


class Oto:
    '''oto.iniのデータを扱います。
    '''
    _values: dict = {}
    _datas_by_file: dict = {}

    def __init__(self, dirpath: str = ""):
        '''
        Parameters
        ----------
        dirpath: str, default ""
            音源のルートディレクトリのパス。引数が与えられている場合self.load(dirpath)を実行する。

        Raises
        ------
        FileNotFoundError
            load実行時oto.iniが見つからなかった場合
        UnicodeDecodeError
            load実行時ファイルがcp932でもutf-8でもなかった場合
        '''

        self._values = {}
        self._datas_by_file = {}
        if dirpath != "":
            self.load(dirpath)

    def files(self) -> int:
        '''
        読み込んだoto.iniファイルの数を返す

        Return
        ------
        len(self._datas_by_file): int
        '''
        return len(self._datas_by_file)

    def records(self) -> int:
        '''
        読み込んだoto.iniの合計行数を返す

        Return
        ------
        len(self._values): int
        '''
        return len(self._values)

    def __getitem__(self, key) -> OtoRecord:
        return self._values[key]

    def haskey(self, key) -> bool:
        return key in self._values

    def load(self, dirpath: str, recursive: bool = False, subdir: str = ""):
        '''
        | dirpathおよびその子フォルダのoto.iniを読み込みself._datas_by_fileとself._valuesを更新する。
        | self._datas_by_fileはdirpathからの相対パスをキーに持つ辞書で、OtoRecord形式のlistを格納する。
        | self._valuesはaliasをkeyとする辞書で、self._datas_by_fileを参照する。
        | aliasの衝突が起きた場合、ファイル名が若いものを参照する。
        | 同一ファイル内でaliasの衝突が起きた場合、offset値の小さいものを参照する。


        Parameters
        ----------
        dirpath: str
            音源のルートディレクトリのパス

        recursive: bool, default False
            | Trueにすると孫フォルダ以下のoto.iniも探索する。

        subdir: str, default ""
            音源ルートディレクトリからの相対パス

        Raises
        ------
        UnicodeDecodeError
            ファイルがcp932でもutf-8でもなかった場合

        '''

        if os.path.isfile(os.path.join(dirpath, subdir, "oto.ini")):
            self._loadFile(os.path.join(dirpath, subdir, "oto.ini"), subdir)
        if subdir == "" or recursive:
            for filename in os.listdir(os.path.join(dirpath, subdir)):
                if os.path.isdir(os.path.join(dirpath, subdir, filename)):
                    self.load(dirpath, recursive, os.path.join(subdir, filename))
                    
    def _setValue(self, alias: str, record: OtoRecord):
        '''
        | self._values[alias]に値をセットする。
        | self._values[alias]が既に存在する場合、ファイル名が若いものをセットする。
        | ファイル名も同じの場合、offsetが小さいものをセットする。

        Parameters
        ----------
        alias: str
            self._valuesのkeyとなるalias

        record: OtoRecord
        '''
        
        if alias not in self._values:
            self._values[alias]=record
        elif record.filename == self._values[alias].filename and record.offset < self._values[alias].offset:
            self._values[alias]=record
        elif record.filename < self._values[alias].filename:
            self._values[alias]=record

    def _loadFile(self, otopath: str, subdir: str):
        '''
        | otopathのoto.iniを読み込みself._datas_by_fileとself._valuesを更新する。
        | ファイルの存在は事前に確認されていること
        
        Parameters
        ----------
        otopath: str
            音源のルートディレクトリのパス

        subdir: str, default ""
            音源ルートディレクトリからの相対パス

        Raises
        ------
        UnicodeDecodeError
            ファイルがcp932でもutf-8でもなかった場合
        '''
        try:
            with open(otopath, "r", encoding="cp932") as fr:
                lines=fr.read().replace("\r", "").split("\n")
        except:
            try:
                with open(otopath, "r", encoding="utf-8") as fr:
                    lines=fr.read().replace("\r", "").split("\n")
            except UnicodeDecodeError as e:
                e.reason="can't read {}. because required character encoding is utf-8 or cp932".format(otopath)
                raise e
        self._datas_by_file[subdir]=[]
        for line in lines:
            if line == "":
                continue
            if "=" not in line:
                continue
            if "," not in line:
                continue
            filename, param=line.split("=")
            params=param.split(",")
            if len(params) != 6:
                continue
            self._datas_by_file[subdir].append(OtoRecord(subdir, filename, params[0], params[1], params[4], params[5], params[2], params[3]))
            self._setValue(params[0], self._datas_by_file[subdir][-1])
            self._setValue(".".join(os.path.join(subdir, filename).split(".")[:-1]), self._datas_by_file[subdir][-1])




import os
import os.path
import locale
import traceback
from logging import Logger

from .Note import Note
from ..settings import logger as mylogger


default_logger = mylogger.get_logger(__name__, False)


class Ust:
    '''
    UTAU sequence textファイルを扱います。

    Attributes
    ----------
    filepath: str
        ustのファイルパス

    version: float, default 1.2
        ustのバージョン。

    project_name: str
        プロジェクト名

    voice_dir: str
        | 音源フォルダへのパス。
        | %voice%はsettings.settings.VOICE_ROOTのこと。

    cache_dir: str
        | wavキャッシュの保存先

    output_file: string
        書き出すwavファイルのパス

    tempo: float
        プロジェクトのbpm

    wavtool: str
        wavtoolのパス

    resamp: str
        resamplerのパス

    flags: str
        デフォルトのフラグ

    mode2: bool, default False
        ピッチのmode2が有効かどうか

    utf8: bool, default False
        ustがutf8で保存されているかどうか

    notes: List of Note
        Noteの配列
    '''

    filepath: str
    _version: float = 1.2
    project_name: str = ""
    voice_dir: str = ""
    cache_dir: str = ""
    output_file: str = ""
    tempo: float = 120.0
    wavtool: str = ""
    resamp: str = ""
    flags: str = ""
    mode2: bool = False
    utf8: bool = False
    notes: list = []

    @property
    def version(self) -> float:
        return self._version

    def __init__(self, filepath: str, *, logger: Logger = None):
        self.logger = logger or default_logger
        self.filepath = filepath
        self.notes = []

    def load(self, filepath: str = ""):
        '''
        self.filepathもしくはfilepathのファイルを読み込みます。

        Parameters
        ----------
        filepath: str, default ""
            読み込むファイルのパス。値が与えられた場合、self.filepathを更新します。

        Raises
        ------
        FileNotFoundError
            self.filepathのファイルが見つからなかった場合
        '''
        if filepath != "":
            self.filepath = filepath

        if not os.path.isfile(self.filepath):
            self.logger.error("{} is not found".format(self.filepath))
            raise FileNotFoundError("{} is not found".format(self.filepath))
        self.logger.info("{} is found. loading file.".format(self.filepath))
        with open(self.filepath, "rb") as fr:
            data = fr.read()
        seek: int = self._load_header(data)
        self.logger.info("loading header complete.")
        self._load_note(data[seek:])
        self.logger.info("loading note complete.notes:{}".format(len(self.notes)))
        for i in range(len(self.notes)):
            if i != 0:
                self.notes[i].prev = self.notes[i - 1]
                self.notes[i - 1].next = self.notes[i]

    def _load_header(self, data: bytes) -> int:
        '''
        | dataをシステムの文字コードもしくはcp932でデコードを試み、各パラメータを更新します。

        Parameters
        ----------
        data: byte
            self.filepathをバイナリで読み込んだもの

        Returns
        -------
        cursor: int
            ヘッダ末尾の位置

        Raises
        ------
        UnicodeDecodeError
            ファイルがシステム既定でもcp932でもデコードできなかった場合
        '''
        version_cursor: int = data.find(b"[#VERSION]")
        setting_cursor: int = data.find(b"[#SETTING]")
        cursor: int = 0
        header_lines: list
        if (setting_cursor == -1 and version_cursor == -1):
            setting_cursor = 0
        cursor = data[setting_cursor + 1:].find(b"[#")
        cursor += setting_cursor
        try:
            header_lines = data[:cursor].decode(locale.getlocale()[1]).replace("\r", "").split("\n")
        except:
            try:
                header_lines = data[:cursor].decode("cp932").replace("\r", "").split("\n")
            except UnicodeDecodeError as e:
                self.logger.error("can't read {}'s header. because required character encoding is system default or cp932".format(self.filepath))
                e.reason = "can't read {}'s header. because required character encoding is system default or cp932".format(self.filepath)
                raise e
        self._load_head_helper(header_lines)
        return cursor

    def _load_head_helper(self, lines):
        vflag: bool = False
        for line in lines:
            if line == "[#VERSION]":
                vflag = True
            elif vflag:
                vflag = False
                self._version = float(line.replace("UST Version", ""))
            elif line.startswith("Tempo="):
                self.tempo = float(line.replace("Tempo=", ""))
            elif line.startswith("Project="):
                self.project_name = line.replace("Project=", "")
            elif line.startswith("VoiceDir="):
                self.voice_dir = line.replace("VoiceDir=", "")
            elif line.startswith("OutFile="):
                self.output_file = line.replace("OutFile=", "")
            elif line.startswith("CacheDir="):
                self.cache_dir = line.replace("CacheDir=", "")
            elif line.startswith("Tool1="):
                self.wavtool = line.replace("Tool1=", "")
            elif line.startswith("Tool2="):
                self.resamp = line.replace("Tool2=", "")
            elif line.startswith("Flags="):
                self.flags = line.replace("Flags=", "")
            elif line.startswith("Mode2="):
                self.mode2 = bool(line.replace("Mode2=", ""))

    def _load_note(self, data: bytes):
        '''
        | dataをcp932もしくはutf-8読み込み、各ノートのパラメータを更新します。

        Parameters
        ----------
        data: byte
            self.filepathをバイナリで読み込み、ヘッダ部分を除いたもの

        Raises
        ------
        UnicodeDecodeError
            ファイルがcp932でもutf-8でも開けなかった場合
        '''
        lines: list = []
        try:
            lines = data.decode("cp932").replace("\r", "").split("\n")
        except:
            try:
                lines = data.decode("utf-8").replace("\r", "").split("\n")
            except UnicodeDecodeError as e:
                self.logger.error("can't read {}'s body. because required character encoding is cp932 or utf-8".format(self.filepath))
                e.reason = "can't read {}'s body. because required character encoding is cp932 or utf-8".format(self.filepath)
                raise e
        self._load_note_helper(lines)

    def _load_note_helper(self, lines: list):
        tempo: float = self.tempo
        note: Note = None
        for line in lines:
            if line == "[#TRACKEND]":
                continue
            elif line.startswith("[#"):
                self.notes.append(Note())
                note = self.notes[-1]
                note.num.init(line.replace("[", "").replace("]", ""))
                note.tempo.init(tempo)
                note.tempo.hasValue = False
                note.flags.init(self.flags)
                note.flags.hasValue = False
            elif line.startswith("Length"):
                try:
                    note.length.init(line.replace("Length=", ""))
                except Exception as e:
                    note.length.init(480)
                    self.logger.warn("{} length can't init. because {}".format(note.num.value,
                                                                               traceback.format_exception_only(type(e), e)[0].rstrip('\n')))
            elif line.startswith("Lyric"):
                try:
                    note.lyric.init(line.replace("Lyric=", ""))
                except Exception as e:
                    note.lyric.init("あ")
                    self.logger.warn("{} lyric can't init. because {}".format(note.num.value,
                                                                              traceback.format_exception_only(type(e), e)[0].rstrip('\n')))
            elif line.startswith("NoteNum"):
                try:
                    note.notenum.init(line.replace("NoteNum=", ""))
                except Exception as e:
                    note.notenum.init(60)
                    self.logger.warn("{} notenum can't init. because {}".format(note.num.value,
                                                                                traceback.format_exception_only(type(e), e)[0].rstrip('\n')))
            elif line.startswith("Tempo"):
                try:
                    note.tempo.init(line.replace("Tempo=", ""))
                    tempo = note.tempo.value
                except Exception as e:
                    self.logger.warn("{} tempo can't init. because {}".format(note.num.value,
                                                                              traceback.format_exception_only(type(e), e)[0].rstrip('\n')))
            elif line.startswith("PreUtterance"):
                try:
                    if line.replace("PreUtterance=", "") != "":
                        note.pre.init(line.replace("PreUtterance=", ""))
                except Exception as e:
                    self.logger.warn("{} pre can't init. because {}".format(note.num.value,
                                                                            traceback.format_exception_only(type(e), e)[0].rstrip('\n')))
            elif line.startswith("@preuttr"):
                try:
                    note.atPre.init(line.replace("@preuttr=", ""))
                except Exception as e:
                    self.logger.warn("{} @preuttr can't init. because {}".format(note.num.value,
                                                                                 traceback.format_exception_only(type(e), e)[0].rstrip('\n')))
            elif line.startswith("VoiceOverlap"):
                try:
                    note.ove.init(line.replace("VoiceOverlap=", ""))
                except Exception as e:
                    self.logger.warn("{} ove can't init. because {}".format(note.num.value,
                                                                            traceback.format_exception_only(type(e), e)[0].rstrip('\n')))
            elif line.startswith("@overlap"):
                try:
                    note.atOve.init(line.replace("@overlap=", ""))
                except Exception as e:
                    self.logger.warn("{} @overlap can't init. because {}".format(note.num.value,
                                                                                 traceback.format_exception_only(type(e), e)[0].rstrip('\n')))
            elif line.startswith("StartPoint"):
                try:
                    note.stp.init(line.replace("StartPoint=", ""))
                except Exception as e:
                    self.logger.warn("{} stp can't init. because {}".format(note.num.value,
                                                                            traceback.format_exception_only(type(e), e)[0].rstrip('\n')))
            elif line.startswith("@stpoint"):
                try:
                    note.atStp.init(line.replace("@stpoint=", ""))
                except Exception as e:
                    self.logger.warn("{} @stpoint can't init. because {}".format(note.num.value,
                                                                                 traceback.format_exception_only(type(e), e)[0].rstrip('\n')))
            elif line.startswith("@filename"):
                try:
                    note.atFileName.init(line.replace("@filename=", ""))
                except Exception as e:
                    self.logger.warn("{} @filename can't init. because {}".format(note.num.value,
                                                                                  traceback.format_exception_only(type(e), e)[0].rstrip('\n')))
            elif line.startswith("@alias"):
                try:
                    note.atAlias.init(line.replace("@alias=", ""))
                except Exception as e:
                    self.logger.warn("{} @alias can't init. because {}".format(note.num.value,
                                                                               traceback.format_exception_only(type(e), e)[0].rstrip('\n')))
            elif line.startswith("Velocity"):
                try:
                    note.velocity.init(line.replace("Velocity=", ""))
                except Exception as e:
                    self.logger.warn("{} Velocity can't init. because {}".format(note.num.value,
                                                                                 traceback.format_exception_only(type(e), e)[0].rstrip('\n')))
            elif line.startswith("Intensity"):
                try:
                    note.intensity.init(line.replace("Intensity=", ""))
                except Exception as e:
                    self.logger.warn("{} Intensity can't init. because {}".format(note.num.value,
                                                                                  traceback.format_exception_only(type(e), e)[0].rstrip('\n')))
            elif line.startswith("Modulation"):
                try:
                    note.modulation.init(line.replace("Modulation=", ""))
                except Exception as e:
                    self.logger.warn("{} Modulation can't init. because {}".format(note.num.value,
                                                                                   traceback.format_exception_only(type(e), e)[0].rstrip('\n')))
            elif line.startswith("PitchBend="):
                try:
                    note.pitches.init_from_str(line.replace("PitchBend=", ""))
                except Exception as e:
                    self.logger.warn("{} PitchBend can't init. because {}".format(note.num.value,
                                                                                traceback.format_exception_only(type(e), e)[0].rstrip('\n')))
            elif line.startswith("PBStart"):
                try:
                    note.pbStart.init(line.replace("PBStart=", ""))
                except Exception as e:
                    self.logger.warn("{} PBStart can't init. because {}".format(note.num.value,
                                                                                traceback.format_exception_only(type(e), e)[0].rstrip('\n')))
            elif line.startswith("PBS"):
                try:
                    note.pbs.init(line.replace("PBS=", ""))
                except Exception as e:
                    self.logger.warn("{} PBS can't init. because {}".format(note.num.value,
                                                                            traceback.format_exception_only(type(e), e)[0].rstrip('\n')))
            elif line.startswith("PBY"):
                try:
                    note.pby.init_from_str(line.replace("PBY=", ""))
                except Exception as e:
                    self.logger.warn("{} PBY can't init. because {}".format(note.num.value,
                                                                            traceback.format_exception_only(type(e), e)[0].rstrip('\n')))
            elif line.startswith("PBM"):
                try:
                    note.pbm.init_from_str(line.replace("PBM=", ""))
                except Exception as e:
                    self.logger.warn("{} PBM can't init. because {}".format(note.num.value,
                                                                            traceback.format_exception_only(type(e), e)[0].rstrip('\n')))
            elif line.startswith("PBW"):
                try:
                    note.pbw.init_from_str(line.replace("PBW=", ""))
                except Exception as e:
                    self.logger.warn("{} PBW can't init. because {}".format(note.num.value,
                                                                            traceback.format_exception_only(type(e), e)[0].rstrip('\n')))
            elif line.startswith("Flags"):
                try:
                    note.flags.init(line.replace("Flags=", ""))
                except Exception as e:
                    self.logger.warn("{} Flags can't init. because {}".format(note.num.value,
                                                                              traceback.format_exception_only(type(e), e)[0].rstrip('\n')))
            elif line.startswith("VBR"):
                try:
                    note.vibrato.init(line.replace("VBR=", ""))
                except Exception as e:
                    self.logger.warn("{} VBR can't init. because {}".format(note.num.value,
                                                                            traceback.format_exception_only(type(e), e)[0].rstrip('\n')))
            elif line.startswith("Envelope"):
                try:
                    note.envelope.init(line.replace("Envelope=", ""))
                except Exception as e:
                    self.logger.warn("{} Envelope can't init. because {}".format(note.num.value,
                                                                                 traceback.format_exception_only(type(e), e)[0].rstrip('\n')))
            elif line.startswith("Label"):
                try:
                    note.label.init(line.replace("Label=", ""))
                except Exception as e:
                    self.logger.warn("{} Label can't init. because {}".format(note.num.value,
                                                                              traceback.format_exception_only(type(e), e)[0].rstrip('\n')))
            elif line.startswith("$direct"):
                try:
                    note.direct.init(line.replace("$direct=", ""))
                except Exception as e:
                    self.logger.warn("{} $direct can't init. because {}".format(note.num.value,
                                                                                traceback.format_exception_only(type(e), e)[0].rstrip('\n')))
            elif line.startswith("$region="):
                try:
                    note.region.init(line.replace("$region=", ""))
                except Exception as e:
                    self.logger.warn("{} $region can't init. because {}".format(note.num.value,
                                                                                traceback.format_exception_only(type(e), e)[0].rstrip('\n')))
            elif line.startswith("$region_end="):
                try:
                    note.region_end.init(line.replace("$region_end=", ""))
                except Exception as e:
                    self.logger.warn("{} $region_end can't init. because {}".format(note.num.value,
                                                                                    traceback.format_exception_only(type(e), e)[0].rstrip('\n')))

    def save(self, filepath: str = "", encoding: str = "cp932"):
        '''
        | self.filepathもしくはfilepathにファイルを保存します。
        | windows版UTAUとの互換性を優先してcp932を優先します。

        Parameters
        ----------
        filepath: str, default ""
        encoding: str, default "cp932"
        '''
        if filepath != "":
            self.filepath = filepath
        if os.path.split(self.filepath)[0] != "":
            os.makedirs(os.path.split(self.filepath)[0], exist_ok=True)
        self.logger.info("saving ust to:{}".format(self.filepath))
        with open(self.filepath, "w", encoding=encoding) as fw:
            fw.write("[#VERSION]\n")
            fw.write("UST Version{:.1f}\n".format(self.version))
            fw.write("[#SETTING]\n")
            fw.write("Tempo={:.2f}\n".format(self.tempo))
            fw.write("Tracks=1\n")
            fw.write("Project={}\n".format(self.project_name))
            fw.write("VoiceDir={}\n".format(self.voice_dir))
            fw.write("OutFile={}\n".format(self.output_file))
            fw.write("CacheDir={}\n".format(self.cache_dir))
            fw.write("Tool1={}\n".format(self.wavtool))
            fw.write("Tool2={}\n".format(self.resamp))
            fw.write("Flags={}\n".format(self.flags))
            fw.write("Mode2={}\n".format(self.mode2))
            for note in self.notes:
                fw.write("[{}]\n".format(str(note.num)))
                fw.write("Length={}\n".format(str(note.length)))
                fw.write("Lyric={}\n".format(str(note.lyric)))
                fw.write("NoteNum={}\n".format(str(note.notenum.value)))
                if note.tempo.hasValue:
                    fw.write("Tempo={}\n".format(str(note.tempo)))
                if note.pre.hasValue:
                    fw.write("PreUtterance={}\n".format(str(note.pre)))
                else:
                    fw.write("PreUtterance=\n")
                if note.ove.hasValue:
                    fw.write("VoiceOverlap={}\n".format(str(note.ove)))
                if note.stp.hasValue:
                    fw.write("StartPoint={}\n".format(str(note.stp)))
                if note.velocity.hasValue:
                    fw.write("Velocity={}\n".format(str(note.velocity)))
                if note.intensity.hasValue:
                    fw.write("Intensity={}\n".format(str(note.intensity)))
                if note.modulation.hasValue:
                    fw.write("Modulation={}\n".format(str(note.modulation)))
                if note.pitches.hasValue:
                    fw.write("PitchBend={}\n".format(str(note.pitches)))
                if note.pbStart.hasValue:
                    fw.write("PBStart={}\n".format(str(note.pbStart)))
                if note.pbs.hasValue:
                    fw.write("PBS={}\n".format(str(note.pbs)))
                if note.pby.hasValue:
                    fw.write("PBY={}\n".format(str(note.pby)))
                if note.pbm.hasValue:
                    fw.write("PBM={}\n".format(str(note.pbm)))
                if note.pbw.hasValue:
                    fw.write("PBW={}\n".format(str(note.pbw)))
                if note.flags.hasValue:
                    fw.write("Flags={}\n".format(str(note.flags)))
                if note.vibrato.hasValue:
                    fw.write("VBR={}\n".format(str(note.vibrato)))
                if note.envelope.hasValue:
                    fw.write("Envelope={}\n".format(str(note.envelope)))
                if note.label.hasValue:
                    fw.write("Label={}\n".format(str(note.label)))
                if note.direct.hasValue:
                    fw.write("$direct={}\n".format(str(note.direct)))
                if note.region.hasValue:
                    fw.write("$region={}\n".format(str(note.region)))
                if note.region_end.hasValue:
                    fw.write("$region_end={}\n".format(str(note.region_end)))
            fw.write("[#TRACKEND]\n")
        self.logger.info("saving ust to:{} complete".format(self.filepath))



import logging

def get_logger(module: str, verbose: bool) -> logging.Logger:
    logger = logging.getLogger(module)
    logger = _set_handler(logger, logging.StreamHandler(), False)
    logger = _set_handler(logger, logging.FileHandler("log.txt", encoding="utf-8"), verbose)
    logger.setLevel(logging.DEBUG)
    logger.propagate = False
    return logger


def _set_handler(logger: logging.Logger, handler: logging.Handler, verbose: bool) -> logging.Logger:
    if verbose:
        handler.setLevel(logging.DEBUG)
    else:
        handler.setLevel(logging.INFO)
    handler.setFormatter(logging.Formatter('%(asctime)s %(name)s:%(lineno)s %(funcName)s [%(levelname)s]: %(message)s'))
    logger.addHandler(handler)
    return logger



'''EntryBase
各ノートパラメータ設定用のベースクラスを定義します。
'''


class EntryBase:
    '''
    エントリー用のベースクラスです。
    継承して使います。
    '''
    _isUpdate: bool = False
    _hasValue: bool = False

    def _set_update(self):
        self._isUpdate = True

    @property
    def isUpdate(self) -> bool:
        return self._isUpdate

    @property
    def hasValue(self) -> bool:
        return self._hasValue


class StringEntry(EntryBase):
    '''
    Str型のvalueをもつエントリー用のベースクラスです。
    継承して使います。
    '''
    _value: str = ""

    @property
    def value(self) -> str:
        return self._value

    @value.setter
    def value(self, value: str):
        self._value = str(value)
        self._set_update()
        self._hasValue = True

    def init(self, value: str):
        self._value = value
        self._hasValue = True

    def __str__(self) -> str:
        return self.value


class IntEntry(EntryBase):
    '''
    int型のvalueをもつエントリー用のベースクラスです。
    継承して使います。
    '''
    _value: int = 0

    @property
    def value(self) -> int:
        return self._value

    @value.setter
    def value(self, value: int):
        try:
            self._value = int(value)
            self._set_update()
            self._hasValue = True
        except:
            raise ValueError("{} is not int".format(value))

    def init(self, value: int):
        try:
            self._value = int(value)
            self._hasValue = True
        except:
            raise ValueError("{} is not int".format(value))

    def __str__(self) -> str:
        return str(self.value)


class FloatEntry(EntryBase):
    '''
    float型のvalueをもつエントリー用のベースクラスです。
    継承して使います。
    '''
    _value: float = 0.0
    point: int = 3

    @property
    def value(self) -> float:
        return self._value

    @value.setter
    def value(self, value: float):
        try:
            self._value = float(value)
            self._set_update()
            self._hasValue = True
        except:
            raise ValueError("{} is not float".format(value))

    def init(self, value: float):
        try:
            self._value = float(value)
            self._hasValue = True
        except:
            raise ValueError("{} is not float".format(value))

    def __str__(self) -> str:
        return ("{:." + str(self.point) + "f}").format(self._value)


class BoolEntry(EntryBase):
    '''
    bool型のvalueをもつエントリー用のベースクラスです。
    継承して使います。
    '''
    _value: bool = False

    @property
    def value(self) -> bool:
        return bool(self._value)

    @value.setter
    def value(self, value: bool):
        self._value = bool(value)
        self._set_update()
        self._hasValue = True

    def init(self, value: bool):
        self._value = value
        self._hasValue = True

    def __str__(self) -> str:
        return str(self.value)


class ListEntry(EntryBase):
    '''
    list型のvalueをもつエントリー用のベースクラスです。
    継承して使います。
    各パラメータのフォーマットが適切かは、self._checl_valueを継承して定義します。
    '''
    _value: list = []
    separater: str = ","

    @property
    def value(self) -> list:
        return self._value

    @value.setter
    def value(self, value: list):
        for v in value:
            self._check_value(v)
        self._value = value
        self._set_update()
        self._hasValue = True

    def _check_value(self, value):
        '''
        Raises
        ------
        ValueError
            値が不適切な時
        '''
        return value

    def init(self, value: list):
        for v in value:
            v = self._check_value(v)
        self._value = value[:]
        self._hasValue = True

    def init_from_str(self, value: str):
        values: list = value.split(self.separater)
        for i in range(len(values)):
            values[i] = self._check_value(values[i])
        self._value = values[:]
        self._hasValue = True

    def __str__(self) -> str:
        return self.separater.join(map(str, self.value))

    def append(self, value):
        value = self._check_value(value)
        self._value.append(value)
        self._isUpdate = True

    def insert(self, pos: int, value):
        value = self._check_value(value)
        self._value.insert(pos, value)
        self._isUpdate = True

    def pop(self, pos: int):
        self._value.pop(pos)
        self._isUpdate = True

    def set(self, pos: int, value):
        value = self._check_value(value)
        self._value[pos] = value
        self._isUpdate = True



import os
import os.path
import shutil
from logging import Logger

import numpy as np
import PyRwu
import PyWavTool

from .Ust import Ust
from .RenderNote import RenderNote
from ..voicebank import VoiceBank
from ..settings import logger as mylogger
from .. import settings

default_logger = mylogger.get_logger(__name__, False)

class FastResamp(PyRwu.Resamp):
    def _getAp(self,
               f0_floor: float,
               f0_ceil: float,
               frame_period: float,
               threshold: float):
        self._ap = np.zeros_like(self._sp)

class Render:
    '''
    ustからwavを生成する処理を扱います。
    '''
    _cache_dir: str
    _output_file: str
    _ust: Ust
    notes: list
    vb: VoiceBank

    def __init__(self, ust: Ust, *, voice_dir: str = "",cache_dir: str = "", output_file: str = "", logger: Logger = None):
        '''
        Parameters
        ----------
        ust: Ust
            load済みのustもしくはUtauPlugin形式のデータ

        voice_dir: str, default ""
            | ustで指定している音源以外で出力する場合に音源のフルパスを指定します。
            | %VOICE%はsettings.VOICE_ROOTの値に置き換えられます。

        cache_dir: str, default ""
            ustで指定している以外の場所にキャッシュファイルを生成する場合、フルパスを指定します。

        output_file: str, default ""
            ustで指定している以外の場所にwavファイルを生成する場合、フルパスを指定します。

        '''
        self.logger = logger or default_logger
        self._ust = ust
        self.notes = []
        
        self.vb = VoiceBank(self._init_voicedir(voice_dir))

        if cache_dir != "":
            self._cache_dir = cache_dir
        else:
            self._cache_dir = ust.cache_dir

        if output_file != "":
            self._output_file = output_file
        else:
            self._output_file = ust.output_file

        for note in ust.notes:
            note.apply_oto(self.vb.oto,self.vb.prefix)

        for note in ust.notes:
            self.notes.append(RenderNote(note, self.vb, self._cache_dir, self._output_file, ust.mode2))

    def _init_voicedir(self, voice_dir) -> str:
        '''
        voice_dirにsettingsの値を適用させ、正しいフルパスを返します。

        Parameters
        ----------
        voice_dir: str, default ""
            ustで指定している音源以外で出力する場合に音源のフルパスを指定します。

        Raises
        ------
        FileNotFoundError
            %VOICE%をsetting.VOICE_ROOTに置き換えたとき、指定したパスのUTAU音源フォルダが見つからなかったとき。

        '''
        if voice_dir == "":
            voice_dir = self._ust.voice_dir
        for root_dir in settings.VOICE_ROOT:
            if VoiceBank.is_utau_voicebank(os.path.join(root_dir,voice_dir.replace("%VOICE%",""))):
                voice_dir = os.path.join(root_dir, voice_dir.replace("%VOICE%",""))
                break
        else:
            self.logger.error("{} is not found or not utau vb.".format(voice_dir))
            raise FileNotFoundError("{} is not found or not utau vb.".format(voice_dir))

        return voice_dir

    def resamp(self, * , force:bool = False):
        '''
        PyRwu.Resampを使用してキャッシュファイルを生成する。

        Parameters
        ----------
        force: bool, default False
            Trueの場合、キャッシュファイルがあっても生成する。
        '''
        os.makedirs(self._cache_dir, exist_ok=True)
        for note in self.notes:
            if not note.require_resamp:
                continue
            if force or not os.path.isfile(note.cache_path):
                self.logger.info("{} {} {} {} {} {} {} {} {} {} {} {} {}".format(note.input_path, note.cache_path, note.target_tone, note.velocity, note.flags,
                                      note.offset, note.target_ms, note.fixed_ms, note.end_ms, note.intensity,
                                      note.modulation, note.tempo, note.pitchbend))
                resamp = PyRwu.Resamp(note.input_path, note.cache_path, note.target_tone, note.velocity, note.flags,
                                      note.offset, note.target_ms, note.fixed_ms, note.end_ms, note.intensity,
                                      note.modulation, note.tempo, note.pitchbend, logger=self.logger)
                resamp.resamp()
            else:
                self.logger.info("{} have be cached".format(note.cache_path))

                

    def fast_resamp(self, * , force:bool = False):
        '''
        FastResampを使用してキャッシュファイルを生成する。

        Parameters
        ----------
        force: bool, default False
            Trueの場合、キャッシュファイルがあっても生成する。
        '''
        os.makedirs(self._cache_dir, exist_ok=True)
        for note in self.notes:
            if not note.require_resamp:
                continue
            if force or not os.path.isfile(note.cache_path):
                self.logger.info("{} {} {} {} {} {} {} {} {} {} {} {} {}".format(note.input_path, note.cache_path, note.target_tone, note.velocity, note.flags,
                                      note.offset, note.target_ms, note.fixed_ms, note.end_ms, note.intensity,
                                      note.modulation, note.tempo, note.pitchbend))
                resamp = FastResamp(note.input_path, note.cache_path, note.target_tone, note.velocity, note.flags,
                                      note.offset, note.target_ms, note.fixed_ms, note.end_ms, note.intensity,
                                      note.modulation, note.tempo, note.pitchbend, logger=self.logger)
                resamp.resamp()
            else:
                self.logger.info("{} have be cached".format(note.cache_path))

    def append(self):
        '''
        PyWavToolを使用してキャッシュファイルから出力ファイルを合成する。
        '''
        output_dir: str = os.path.split(self._output_file)[0]
        if output_dir != "":
            os.makedirs(output_dir, exists_ok=True)

        if os.path.isfile(self._output_file+".whd"):
            os.remove(self._output_file+".whd")

        if os.path.isfile(self._output_file+".dat"):
            os.remove(self._output_file+".dat")
        wavtool = PyWavTool.WavTool(self._output_file)
        for note in self.notes:
            if note.direct:
                self.logger.info("{} {} {} {}".format(note.input_path, note.envelope, note.stp+note.offset, note.output_ms))
                wavtool.inputCheck(note.input_path)
                wavtool.setEnvelope([float(item) for item in note.envelope.split(" ")])
                wavtool.applyData(note.stp + note.offset, note.output_ms)
            else:
                self.logger.info("{} {} {} {}".format(note.cache_path, note.envelope, note.stp, note.output_ms))
                wavtool.inputCheck(note.cache_path)
                wavtool.setEnvelope([float(item) for item in note.envelope.split(" ")])
                wavtool.applyData(note.stp, note.output_ms)
        wavtool.write()
            
        with open(self._output_file, "wb") as fw:
            with open(self._output_file+".whd", "rb") as fr:
                fw.write(fr.read())
            with open(self._output_file+".dat", "rb") as fr:
                fw.write(fr.read())

        if os.path.isfile(self._output_file+".whd"):
            os.remove(self._output_file+".whd")

        if os.path.isfile(self._output_file+".dat"):
            os.remove(self._output_file+".dat")

    def clean(self):
        '''
        self._cache_dirとself._output_fileが存在すれば削除する。
        '''
        if os.path.isdir(self._cache_dir):
            shutil.rmtree(self._cache_dir)
        if os.path.isfile(self._output_file):
            os.remove(self._output_file)